from typing import Optional

from boaviztapi import config
from boaviztapi.dto.component import ComponentDTO
from boaviztapi.dto.usage.usage import mapper_usage, Usage
from boaviztapi.model.component import ComponentPowerSupply, ComponentMotherboard, ComponentCase
from boaviztapi.service.archetype import get_component_archetype


class PowerSupply(ComponentDTO):
    unit_weight: Optional[float] = None


class Motherboard(ComponentDTO):
    pass


class Case(ComponentDTO):
    case_type: str = None


def mapper_power_supply(power_supply_dto: PowerSupply, archetype=get_component_archetype(config["default_power_supply"], "power_supply")) -> ComponentPowerSupply:
    power_supply_component = ComponentPowerSupply(archetype=archetype)
    power_supply_component.usage = mapper_usage(power_supply_dto.usage or Usage(), archetype=archetype.get("USAGE"))

    if power_supply_dto.units is not None:
        power_supply_component.units.set_input(power_supply_dto.units)

    if power_supply_dto.unit_weight is not None:
        power_supply_component.unit_weight.set_input(power_supply_dto.unit_weight)

    return power_supply_component


def mapper_motherboard(motherboard_dto: Motherboard) -> ComponentMotherboard:
    motherboard_component = ComponentMotherboard()
    motherboard_component.usage = mapper_usage(motherboard_dto.usage or Usage())

    if motherboard_dto.units is not None:
        motherboard_component.units.set_input(motherboard_dto.units)

    return motherboard_component


def mapper_case(case_dto: Case, archetype=get_component_archetype(config["default_case"], "case")) -> ComponentCase:
    case_component = ComponentCase(archetype=archetype)
    case_component.usage = mapper_usage(case_dto.usage or Usage(), archetype=archetype.get("USAGE"))

    if case_dto.units is not None:
        case_component.units.set_input(case_dto.units)

    if case_dto.case_type is not None:
        case_component.case_type.set_input(case_dto.case_type)

    return case_component



from typing import Tuple

NumberSignificantFigures = Tuple[float, int]


def allocate(impact, duration, life_time) -> [NumberSignificantFigures, NumberSignificantFigures, NumberSignificantFigures]:

    if duration > life_time.value:
        allocation_ratio, allocation_ratio_min, allocation_ratio_max = 1, 1, 1
    else:
        allocation_ratio = duration / life_time.value
        allocation_ratio_min = duration / life_time.max
        allocation_ratio_max = duration / life_time.min

    return impact.value * allocation_ratio, impact.min * allocation_ratio_min, impact.max * allocation_ratio_max


from abc import abstractmethod

import boaviztapi.utils.roundit as rd
from boaviztapi.model import ComputedImpacts
from boaviztapi.model.boattribute import Boattribute
from boaviztapi.model.usage import ModelUsage
from boaviztapi.service.archetype import get_arch_value, get_arch_component


class Component:
    NAME = "COMPONENT"

    def __init__(self, archetype=None, **kwargs):
        self.impact_factor = {}
        self.archetype = archetype
        self.units = Boattribute(
            default=get_arch_value(archetype, 'units', 'default', default=1),
            min=get_arch_value(archetype, 'units', 'min'),
            max=get_arch_value(archetype, 'units', 'max')
        )
        self._usage = None

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    @property
    def usage(self) -> ModelUsage:
        if self._usage is None:
            self._usage = ModelUsage(archetype=get_arch_component(self.archetype,"USAGE"))
        return self._usage

    @usage.setter
    def usage(self, value: int) -> None:
        self._usage = value

    def impact_use(self, impact_type: str, duration:float) -> ComputedImpacts:
        if not self.usage.avg_power.is_set():
            raise NotImplementedError
        impact_factor = self.usage.elec_factors[impact_type]

        impacts = impact_factor.value * (self.usage.avg_power.value / 1000) * self.usage.use_time_ratio.value * duration
        sig_fig = self.__compute_significant_numbers_usage(impact_factor.value)

        max_impact = impact_factor.max * (self.usage.avg_power.max / 1000) * self.usage.use_time_ratio.min * duration
        min_impact = impact_factor.min * (self.usage.avg_power.min / 1000) * self.usage.use_time_ratio.max * duration


        return impacts, sig_fig, min_impact, max_impact, []

    def __compute_significant_numbers_usage(self, impact_factor: float) -> int:
        return rd.min_significant_figures(self.usage.avg_power.value, self.usage.use_time.value, impact_factor)

    @abstractmethod
    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        raise NotImplementedError


from typing import Union, Optional

from boaviztapi import config
from boaviztapi.model.component import Component
from boaviztapi.model.device import Device
from boaviztapi.model.impact import Impact, IMPACT_CRITERIAS, IMPACT_PHASES
from boaviztapi.service.allocation import allocate

NOT_IMPLEMENTED = 'not implemented'


def get_model_single_impact(model: Union[Component, Device],
                            phase: str,
                            impact_type: str,
                            duration: Union[int, str] = config["default_duration"]) -> Optional[Impact]:
    try:
        impact_function = model.__getattribute__(f'impact_{phase}')

        if phase == "use":
            impact, significant_figures, min_impact, max_impact, warnings = impact_function(impact_type, duration)
        else:
            impact, significant_figures, min_impact, max_impact, warnings = impact_function(impact_type)
            impact, min_impact, max_impact = allocate(Impact(value=impact, min=min_impact, max=max_impact), duration,
                                                      model.usage.hours_life_time)

        return Impact(
            value=impact * model.units.value,
            significant_figures=significant_figures,
            min=min_impact * model.units.min,
            max=max_impact * model.units.max,
            warnings=list(set(warnings))
        )
    except (AttributeError, NotImplementedError):
        pass


def bottom_up(model: Union[Component, Device], selected_criteria=config["default_criteria"],
              duration=config["default_duration"]) -> dict:
    impacts = {}
    for criteria in IMPACT_CRITERIAS:
        if "all" not in selected_criteria:
            if criteria.name not in selected_criteria:
                continue
        impacts[criteria.name] = {}
        for phase in IMPACT_PHASES:
            single_impact = get_model_single_impact(model, phase, criteria.name, duration=duration)
            impacts[criteria.name][phase] = single_impact.to_json() if single_impact else NOT_IMPLEMENTED
        impacts[criteria.name]["unit"] = criteria.unit
        impacts[criteria.name]["description"] = criteria.description
    return impacts



from dataclasses import dataclass
from typing import Optional

import boaviztapi.utils.roundit as rd

@dataclass
class ImpactCriteria:
    name: str
    unit: str
    description: str
    method: Optional[str] = None
class Impact:
    def __init__(self, **kwargs):
        self.value = 0
        self.significant_figures = 1
        self.min = 0
        self.max = 0
        self.warnings = []

        for attr, val in kwargs.items():
            if val is not None:
                self.__setattr__(attr, val)
    def add_warning(self, warn):
        self.warnings.append(warn)
    def to_json(self):
        json = {"value": rd.round_to_sigfig(self.value, self.significant_figures), "significant_figures": self.significant_figures}
        if self.min or self.min==0: json['min'] = rd.round_to_sigfig(self.min, self.significant_figures)
        if self.max or self.max==0: json['max'] = rd.round_to_sigfig(self.max, self.significant_figures)
        if self.warnings: json['warnings'] = self.warnings

        return json
    def rounded_value(self):
        return rd.round_to_sigfig(self.value, self.significant_figures)


GWP = ImpactCriteria(name="gwp", unit="kgCO2eq", description="Total climate change")
ADP = ImpactCriteria(name="adp", unit="kgSbeq", description="Use of minerals and fossil ressources")
PE = ImpactCriteria(name="pe", unit="MJ", description="Consumption of primary energy")
GWPPb = ImpactCriteria(name="gwppb", unit="kg CO2 eq.", method="PEF", description="Climate change - Contribution of biogenic emissions")
GWPPf = ImpactCriteria(name="gwppf", unit="kg CO2 eq.", method="PEF", description="Climate change - Contribution of fossil fuel emissions")
GWPPlu = ImpactCriteria(name="gwpplu", unit="kg CO2 eq.", method="PEF", description="Climate change - Contribution of emissions from land use change")
IR = ImpactCriteria(name="ir", unit="kg U235 eq.", method="PEF", description="Emissions of radionizing substances")
LU = ImpactCriteria(name="lu", unit="No dimension", method="PEF", description="Land use")
ODP = ImpactCriteria(name="odp", unit="kg CFC-11 eq.", method="PEF", description="Depletion of the ozone layer")
PM = ImpactCriteria(name="pm", unit="Disease occurrence", method="PEF", description="Fine particle emissions")
POCP = ImpactCriteria(name="pocp", unit="kg NMVOC eq.", method="PEF", description="Photochemical ozone formation")
WU = ImpactCriteria(name="wu", unit="m3 eq.", method="PEF", description="Use of water resources")
MIPS = ImpactCriteria(name="mips", unit="kg", description="Material input per unit of service")
ADPe = ImpactCriteria(name="adpe", unit="kg SB eq.", method="PEF", description="Use of mineral and metal resources")
ADPf = ImpactCriteria(name="adpf", unit="MJ", method="PEF", description="Use of fossil resources (including nuclear)")
AP = ImpactCriteria(name="ap", unit="mol H+ eq.", method="PEF", description="Acidification")
CTUe = ImpactCriteria(name="ctue", unit="CTUe", method="PEF", description="Freshwater ecotoxicity")
CTUh_c = ImpactCriteria(name="ctuh_c", unit="CTUh", method="PEF", description="Human Toxicity - Carcinogenic Effects")
CTUh_nc = ImpactCriteria(name="ctuh_nc", unit="CTUh", method="PEF", description="Human toxicity - non-carcinogenic effects")
Epf = ImpactCriteria(name="epf", unit="kg P eq.", method="PEF", description="Eutrophication of freshwater")
Epm = ImpactCriteria(name="epm", unit="kg N eq.", method="PEF", description="Eutrophication of marine waters")
Ept = ImpactCriteria(name="ept", unit="mol N eq.", method="PEF", description="Terrestrial eutrophication")

IMPACT_CRITERIAS = [GWP, ADP, PE, GWPPb, GWPPf, GWPPlu, IR, LU, ODP, PM, POCP, WU, MIPS, ADPe, ADPf, AP, CTUe, CTUh_c, CTUh_nc, Epf, Epm, Ept]
IMPACT_PHASES = ["embedded", "use"]

class ImpactFactor:
    def __init__(self, **kwargs):
        self.value = 0
        self.min = 0
        self.max = 0
        for attr, val in kwargs.items():
            if val is not None:
                self.__setattr__(attr, val)


import os

import numpy as np
import pandas as pd

from boaviztapi import data_dir

_electricity_emission_factors_df = pd.read_csv(os.path.join(data_dir, 'crowdsourcing/electrical_mix.csv'))

def electical_mix_jsonify(electrical_mix):
    json = {"min-max": {}}
    for country_code in electrical_mix.columns:
        if country_code == "english_name" or country_code == "french_name" or country_code == "unit" or country_code == "source" or country_code == "priority" or country_code == "reference_year" or country_code == "name":
            continue
        if country_code not in json:
            json[country_code] = {}
            json[country_code]['country'] = electrical_mix[country_code][0]
        for index, row in electrical_mix.iterrows():
            if row["name"] == "renewable_energy" or row["name"] == "country":
                continue
            if np.isnan(float(row[country_code])):
                continue
            if row["name"] in json[country_code]:
                if row["priority"] > json[country_code][row["name"]]["priority"]:
                    continue

            json[country_code][row["name"]] = {}
            json[country_code][row["name"]]["unit"] = row["unit"]
            json[country_code][row["name"]]["source"] = row["source"]
            json[country_code][row["name"]]["priority"] = row["priority"]
            json[country_code][row["name"]]["reference_year"] = row["reference_year"]
            json[country_code][row["name"]]["value"] = float(row[country_code])

            if row["name"] not in json["min-max"]:
                json["min-max"][row["name"]] = {}
                json["min-max"][row["name"]]["min"] = float(row[country_code])
                json["min-max"][row["name"]]["max"] = float(row[country_code])
            else:
                if float(row[country_code]) < json["min-max"][row["name"]]["min"]:
                    json["min-max"][row["name"]]["min"] = float(row[country_code])
                if float(row[country_code]) > json["min-max"][row["name"]]["max"]:
                    json["min-max"][row["name"]]["max"] = float(row[country_code])

    return {"electricity": json}

if __name__ == '__main__':
    print(electical_mix_jsonify(_electricity_emission_factors_df))


from typing import List, Union, Optional
from fastapi import APIRouter, Query, Body

from boaviztapi import config
from boaviztapi.dto.device.user_terminal import Monitor, UsbStick, ExternalSSD, ExternalHDD
from boaviztapi.routers.openapi_doc.descriptions import all_archetype_user_terminals, all_peripheral_categories, \
    get_archetype_config_desc, peripheral_description
from boaviztapi.routers.openapi_doc.examples import end_user_terminal
from boaviztapi.routers.terminal_router import user_terminal_impact, get_all_archetype_name, get_archetype_config

peripheral_router = APIRouter(
    prefix='/v1/peripheral',
    tags=['peripheral']
)


@peripheral_router.get('/all',
                       description=all_peripheral_categories)
async def peripheral_get_all_categories():
    return {
        "monitor": "v1/peripheral/monitor",
        "usb_stick": "v1/peripheral/usb_stick",
        "external_ssd": "v1/peripheral/external_ssd",
        "external_hdd": "v1/peripheral/external_hdd"
    }


@peripheral_router.get('/monitor/archetypes',
                       description=all_archetype_user_terminals)
async def monitor_get_all_archetype_name():
    return get_all_archetype_name('monitor')


@peripheral_router.get('/monitor/archetype_config',
                       description=get_archetype_config_desc)
async def monitor_get_archetype_config(archetype: str = Query(example=config["default_monitor"])):
    return get_archetype_config(archetype)


@peripheral_router.post('/monitor', description=peripheral_description)
async def monitor_impact(monitor: Monitor = Body(None, example=end_user_terminal),
                         verbose: bool = True,
                         duration: Optional[float] = config["default_duration"],
                         archetype: str = config["default_monitor"],
                         criteria: List[str] = Query(config["default_criteria"])):
    return await user_terminal_impact(user_terminal_dto=monitor,
                                      verbose=verbose,
                                      duration=duration,
                                      criteria=criteria,
                                      archetype=archetype)


@peripheral_router.get('/monitor', description=peripheral_description)
async def monitor_impact(archetype: str = config["default_monitor"],
                         verbose: bool = True,
                         duration: Optional[float] = config["default_duration"],
                         criteria: List[str] = Query(config["default_criteria"])):
    return await user_terminal_impact(user_terminal_dto=Monitor(),
                                      verbose=verbose,
                                      duration=duration,
                                      criteria=criteria,
                                      archetype=archetype)


@peripheral_router.get('/usb_stick/archetypes',
                       description=all_archetype_user_terminals)
async def usb_stick_get_all_archetype_name():
    return get_all_archetype_name('usb_stick')


@peripheral_router.get('/usb_stick/archetype_config',
                       description=get_archetype_config_desc)
async def usb_stick_get_archetype_config(archetype: str = Query(example=config["default_usb_stick"])):
    return get_archetype_config(archetype)


@peripheral_router.post('/usb_stick', description=peripheral_description)
async def usb_stick_impact(usb_stick: UsbStick = Body(None, example=end_user_terminal),
                           verbose: bool = True,
                           duration: Optional[float] = config["default_duration"],
                           archetype: str = config["default_usb_stick"],
                           criteria: List[str] = Query(config["default_criteria"])):
    return await user_terminal_impact(user_terminal_dto=usb_stick,
                                      verbose=verbose,
                                      duration=duration,
                                      criteria=criteria,
                                      archetype=archetype)


@peripheral_router.get('/usb_stick', description=peripheral_description)
async def usb_stick_impact(archetype: str = config["default_usb_stick"],
                           verbose: bool = True,
                           duration: Optional[float] = config["default_duration"],
                           criteria: List[str] = Query(config["default_criteria"])):
    return await user_terminal_impact(user_terminal_dto=UsbStick(),
                                      verbose=verbose,
                                      duration=duration,
                                      criteria=criteria,
                                      archetype=archetype)


@peripheral_router.get('/external_ssd/archetypes',
                       description=all_archetype_user_terminals)
async def external_ssd_get_all_archetype_name():
    return get_all_archetype_name('external_ssd')


@peripheral_router.get('/external_ssd/archetype_config',
                       description=get_archetype_config_desc)
async def external_ssd_get_archetype_config(archetype: str = Query(example=config["default_external_ssd"])):
    return get_archetype_config(archetype)


@peripheral_router.post('/external_ssd', description=peripheral_description)
async def external_ssd_impact(external_ssd: ExternalSSD = Body(None, example=end_user_terminal),
                              verbose: bool = True,
                              duration: Optional[float] = config["default_duration"],
                              archetype: str = config["default_external_ssd"],
                              criteria: List[str] = Query(config["default_criteria"])):
    return await user_terminal_impact(user_terminal_dto=external_ssd,
                                      verbose=verbose,
                                      duration=duration,
                                      criteria=criteria,
                                      archetype=archetype)


@peripheral_router.get('/external_ssd', description=peripheral_description)
async def external_ssd_impact(archetype: str = config["default_external_ssd"],
                              verbose: bool = True,
                              duration: Optional[float] = config["default_duration"],
                              criteria: List[str] = Query(config["default_criteria"])):
    return await user_terminal_impact(user_terminal_dto=ExternalSSD(),
                                      verbose=verbose,
                                      duration=duration,
                                      criteria=criteria,
                                      archetype=archetype)


@peripheral_router.get('/external_hdd/archetypes',
                       description=all_archetype_user_terminals)
async def external_hdd_get_all_archetype_name():
    return get_all_archetype_name('external_hdd')


@peripheral_router.get('/external_hdd/archetype_config',
                       description=get_archetype_config_desc)
async def external_hdd_get_archetype_config(archetype: str = Query(example=config["default_external_hdd"])):
    return get_archetype_config(archetype)


@peripheral_router.post('/external_hdd', description=peripheral_description)
async def external_hdd_impact(external_hdd: ExternalHDD = Body(None, example=end_user_terminal),
                              verbose: bool = True,
                              duration: Optional[float] = config["default_duration"],
                              archetype: str = config["default_external_hdd"],
                              criteria: List[str] = Query(config["default_criteria"])):
    return await user_terminal_impact(user_terminal_dto=external_hdd,
                                      verbose=verbose,
                                      duration=duration,
                                      criteria=criteria,
                                      archetype=archetype)


@peripheral_router.get('/external_hdd', description=peripheral_description)
async def external_hdd_impact(archetype: str = config["default_external_hdd"],
                              verbose: bool = True,
                              duration: Optional[float] = config["default_duration"],
                              criteria: List[str] = Query(config["default_criteria"])):
    return await user_terminal_impact(user_terminal_dto=ExternalHDD(),
                                      verbose=verbose,
                                      duration=duration,
                                      criteria=criteria,
                                      archetype=archetype)



import os
import os

import pandas as pd
from fastapi import APIRouter, Query

from boaviztapi.dto.component.cpu import CPU
from boaviztapi.model import impact
from boaviztapi.model.component import ComponentCase
from boaviztapi.model.component.cpu import attributes_from_cpu_name
from boaviztapi.routers.openapi_doc.descriptions import country_code, cpu_family, cpu_model_range, ssd_manufacturer, \
    ram_manufacturer, case_type, name_to_cpu, cpu_names, impacts_criteria
from boaviztapi.service.factor_provider import get_available_countries

utils_router = APIRouter(
    prefix='/v1/utils',
    tags=['utils']
)

data_dir = os.path.join(os.path.dirname(__file__), '../data')
_cpu_specs = pd.read_csv(os.path.join(data_dir, 'crowdsourcing/cpu_specs.csv'))
_cpu_manuf = pd.read_csv(os.path.join(data_dir, 'crowdsourcing/cpu_manufacture.csv'))
_ssd_manuf = pd.read_csv(os.path.join(data_dir, 'crowdsourcing/ssd_manufacture.csv'))
_ram_manuf = pd.read_csv(os.path.join(data_dir, 'crowdsourcing/ram_manufacture.csv'))


@utils_router.get('/country_code', description=country_code)
async def utils_get_all_countries():
    return get_available_countries()


@utils_router.get('/cpu_family', description=cpu_family)
async def utils_get_all_cpu_family():
    df = _cpu_manuf[_cpu_manuf["family"].notna()]
    return [*df["family"].unique()]


@utils_router.get('/cpu_model_range', description=cpu_model_range)
async def utils_get_all_cpu_model_range():
    df = _cpu_specs[_cpu_specs["model_range"].notna()]
    return [*df["model_range"].unique()]


@utils_router.get('/ssd_manufacturer', description=ssd_manufacturer)
async def utils_get_all_ssd_manufacturer():
    df = _ssd_manuf[_ssd_manuf["manufacturer"].notna()]

    return [*df["manufacturer"].unique()]


@utils_router.get('/ram_manufacturer', description=ram_manufacturer)
async def utils_get_all_ram_manufacturer():
    df = _ram_manuf[_ram_manuf["manufacturer"].notna()]

    return [*df["manufacturer"].unique()]


@utils_router.get('/case_type', description=case_type)
async def utils_get_all_case_type():
    return ComponentCase.AVAILABLE_CASE_TYPE


@utils_router.get('/name_to_cpu', description=name_to_cpu)
async def name_to_cpu(cpu_name: str = Query(example="Intel Core i7-9700K")):
    name, manufacturer, code_name, model_range, tdp, cores, total_die_size, total_die_size_source, source  = attributes_from_cpu_name(cpu_name)
    return CPU(family=code_name, name=name, tdp=tdp, core_units=cores, die_size=total_die_size, model_range=model_range, manufacturer=manufacturer)

@utils_router.get('/cpu_name', description=cpu_names)
async def utils_get_all_cpu_name():
    df = _cpu_specs[_cpu_specs["name"].notna()]
    return [*df["name"].unique()]

@utils_router.get('/impact_criteria', description=impacts_criteria)
async def utils_get_all_impacts_criteria():
    return impact.IMPACT_CRITERIAS



import boaviztapi.utils.roundit as rd
from boaviztapi.model.component.component import Component, ComputedImpacts
from boaviztapi.model.impact import ImpactFactor
from boaviztapi.service.factor_provider import get_impact_factor


class ComponentMotherboard(Component):
    NAME = "MOTHERBOARD"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        impact = ImpactFactor(
            value=get_impact_factor(item='motherboard', impact_type=impact_type)['impact'],
            min=get_impact_factor(item='motherboard', impact_type=impact_type)['impact'],
            max=get_impact_factor(item='motherboard', impact_type=impact_type)['impact']
        )

        significant_figures = rd.min_significant_figures(impact.value)
        return impact.value, significant_figures, impact.min, impact.max, ["End of life is not included in the calculation"]


import os
from typing import Tuple

import pandas as pd

import boaviztapi.utils.roundit as rd
from boaviztapi import config, data_dir
from boaviztapi.model.boattribute import Boattribute
from boaviztapi.model.component.component import Component, ComputedImpacts
from boaviztapi.model.impact import ImpactFactor
from boaviztapi.service.archetype import get_component_archetype, get_arch_value
from boaviztapi.service.factor_provider import get_impact_factor
from boaviztapi.utils.fuzzymatch import fuzzymatch_attr_from_pdf


class ComponentSSD(Component):
    _ssd_df = pd.read_csv(os.path.join(data_dir, 'crowdsourcing/ssd_manufacture.csv'))

    NAME = "SSD"

    __DISK_TYPE = 'ssd'

    DEFAULT_SSD_CAPACITY = 1000
    DEFAULT_SSD_DENSITY = 48.5

    def __init__(self, archetype=get_component_archetype(config["default_ssd"], "ssd"), **kwargs):
        super().__init__(archetype=archetype, **kwargs)
        self.manufacturer = Boattribute(
            default=get_arch_value(archetype, 'manufacturer', 'default'),
            min=get_arch_value(archetype, 'manufacturer', 'min'),
            max=get_arch_value(archetype, 'manufacturer', 'max')
        )
        self.capacity = Boattribute(
            unit="GB",
            default=get_arch_value(archetype, 'capacity', 'default'),
            min=get_arch_value(archetype, 'capacity', 'min'),
            max=get_arch_value(archetype, 'capacity', 'max')
        )
        self.density = Boattribute(
            unit="GB/cm2",
            complete_function=self._complete_density,
            default=get_arch_value(archetype, 'density', 'default'),
            min=get_arch_value(archetype, 'density', 'min'),
            max=get_arch_value(archetype, 'density', 'max')
        )
        self.layers = Boattribute(
            default=get_arch_value(archetype, 'layers', 'default'),
            min=get_arch_value(archetype, 'layers', 'min'),
            max=get_arch_value(archetype, 'layers', 'max')
        )

    # IMPACT CALCUATION
    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        ssd_die_impact, ssd_impact = self.__get_impact_constants(impact_type)
        sign_figures = self.__compute_significant_numbers(ssd_die_impact.value, ssd_impact.value)
        impact = self.__compute_impact_manufacture(ssd_die_impact, ssd_impact)
        return impact.value, sign_figures, impact.min, impact.max, ["End of life is not included in the calculation"]


    def __get_impact_constants(self, impact_type: str) -> Tuple[ImpactFactor, ImpactFactor]:
        ssd_die_impact = ImpactFactor(
            value=get_impact_factor(item='ssd', impact_type=impact_type)['die_impact'],
            min=get_impact_factor(item='ssd', impact_type=impact_type)['die_impact'],
            max=get_impact_factor(item='ssd', impact_type=impact_type)['die_impact']
        )
        ssd_impact = ImpactFactor(
            value=get_impact_factor(item='ssd', impact_type=impact_type)['impact'],
            min=get_impact_factor(item='ssd', impact_type=impact_type)['impact'],
            max=get_impact_factor(item='ssd', impact_type=impact_type)['impact']
        )
        return ssd_die_impact, ssd_impact

    def __compute_significant_numbers(self, ssd_die_impact: float, ssd_impact: float) -> int:
        return rd.min_significant_figures(self.density.value, ssd_die_impact, ssd_impact)

    def __compute_impact_manufacture(self, ssd_die_impact: ImpactFactor, ssd_impact: ImpactFactor) -> ImpactFactor:
        return ImpactFactor(
            value=(self.capacity.value / self.density.value) * ssd_die_impact.value + ssd_impact.value,
            min=(self.capacity.min / self.density.max) * ssd_die_impact.min + ssd_impact.min,
            max=(self.capacity.max / self.density.min) * ssd_die_impact.max + ssd_impact.max
        )

    # COMPLETION
    def _complete_density(self):
        sub = self._ssd_df
        if self.manufacturer.has_value():
            corrected_manufacturer = fuzzymatch_attr_from_pdf(self.manufacturer.value, "manufacturer", sub)
            sub = sub[sub['manufacturer'] == corrected_manufacturer]
            if corrected_manufacturer != self.manufacturer.value:
                self.manufacturer.set_changed(corrected_manufacturer)

        if self.layers.has_value():
            sub = sub[sub['layers'] == self.layers.value]

        if len(sub) == 1:
            self.density.set_completed(float(sub['density']), source=str(sub['source'].iloc[0]),
                                                         min=float(sub['density']), max=float(sub['density']))

        elif (len(sub) == 0 or len(sub) == len(self._ssd_df)) and self.density.has_value():
            return

        else:
            self.density.set_completed(
                float(sub['density'].mean()),
                source="Average of " + str(len(sub)) + " rows",
                min=float(sub['density'].min()),
                max=float(sub['density'].max())
            )


import os
from typing import List, Union, Optional

import pandas as pd

from fastapi import APIRouter, Query, Body, HTTPException

from boaviztapi import config, data_dir
from boaviztapi.dto.device import Cloud
from boaviztapi.dto.device.device import mapper_cloud_instance
from boaviztapi.routers.openapi_doc.descriptions import cloud_provider_description, all_default_cloud_instances, \
    all_default_cloud_providers,get_instance_config
from boaviztapi.routers.openapi_doc.examples import cloud_example
from boaviztapi.routers.server_router import server_impact
from boaviztapi.service.archetype import get_cloud_instance_archetype, get_device_archetype_lst

cloud_router = APIRouter(
    prefix='/v1/cloud',
    tags=['cloud']
)

@cloud_router.get('/instance/instance_config',
                   description=get_instance_config)
async def get_archetype_config(provider: str = Query(config["default_cloud_provider"], example=config["default_cloud_provider"]),
                               instance_type: str = Query(config["default_cloud"], example=config["default_cloud"])):

    result = get_cloud_instance_archetype(instance_type, provider)
    if not result:
        raise HTTPException(status_code=404, detail=f"{instance_type} at {provider} not found")
    return result

@cloud_router.post('/instance',
                   description=cloud_provider_description)
async def instance_cloud_impact(cloud_instance: Cloud = Body(None, example=cloud_example),
                                verbose: bool = True,
                                duration: Optional[float] = config["default_duration"],
                                criteria: List[str] = Query(config["default_criteria"])):
    instance_archetype = get_cloud_instance_archetype(cloud_instance.instance_type, cloud_instance.provider)

    if not instance_archetype:
        raise HTTPException(status_code=404, detail=f"{cloud_instance.instance_type} at {cloud_instance.provider} not found")

    instance_model = mapper_cloud_instance(cloud_instance, archetype=instance_archetype)

    return await server_impact(
        device=instance_model,
        verbose=verbose,
        duration=duration,
        criteria=criteria
    )

@cloud_router.get('/instance',
                   description=cloud_provider_description)
async def instance_cloud_impact(provider: str = Query(config["default_cloud_provider"], example=config["default_cloud_provider"]),
                                instance_type: str = Query(config["default_cloud"], example=config["default_cloud"]), verbose: bool = True,
                                duration: Optional[float] = config["default_duration"],
                                criteria: List[str] = Query(config["default_criteria"])):
    cloud_instance = Cloud()
    cloud_instance.usage = {}
    instance_archetype = get_cloud_instance_archetype(instance_type, provider)

    if not instance_archetype:
        raise HTTPException(status_code=404,
                            detail=f"{cloud_instance.instance_type} at {cloud_instance.provider} not found")

    instance_model = mapper_cloud_instance(cloud_instance, archetype=instance_archetype)

    return await server_impact(
        device=instance_model,
        verbose=verbose,
        duration=duration,
        criteria=criteria
    )

@cloud_router.get('/instance/all_instances',
                  description=all_default_cloud_instances)
async def server_get_all_archetype_name(provider: str = Query(None, example="aws")):
    if not os.path.exists(data_dir +'/archetypes/cloud/' + provider + '.csv'):
        raise HTTPException(status_code=404, detail=f"No available data for this cloud provider ({provider})")
    return get_device_archetype_lst(os.path.join(data_dir, 'archetypes/cloud/' + provider + '.csv'))

@cloud_router.get('/instance/all_providers',
                  description=all_default_cloud_providers)
async def server_get_all_provider_name():
    df = pd.read_csv(os.path.join(data_dir, 'archetypes/cloud/providers.csv'))
    return df['provider.name'].tolist()


from abc import abstractmethod
from typing import Tuple, List

from boaviztapi.model.boattribute import Boattribute
from boaviztapi.model.component import Component
from boaviztapi.model.usage import ModelUsage

ComputedImpacts = Tuple[float, int]


class Device:

    def __init__(self, archetype=None, **kwargs):
        self.impact_factor = {}
        self.archetype = archetype
        self.units = Boattribute(
            default=1,
            min=1,
            max=1
        )
        self._usage = None
        pass

    @property
    def usage(self) -> ModelUsage:
        return self._usage

    @usage.setter
    def usage(self, value: int) -> None:
        self._usage = value

    @property
    def components(self) -> List[Component]:
        return []

    @abstractmethod
    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        raise NotImplementedError

    @abstractmethod
    def  impact_use(self, impact_type: str, duration: float) -> ComputedImpacts:
        raise NotImplementedError

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value




import boaviztapi.utils.roundit as rd
from boaviztapi.model import ComputedImpacts
from boaviztapi.model.component.component import Component
from boaviztapi.model.impact import ImpactFactor
from boaviztapi.service.factor_provider import get_impact_factor


class ComponentAssembly(Component):
    NAME = "ASSEMBLY"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        impact_factor = ImpactFactor(
            value=get_impact_factor(item='assembly', impact_type=impact_type)['impact'],
            min=get_impact_factor(item='assembly', impact_type=impact_type)['impact'],
            max=get_impact_factor(item='assembly', impact_type=impact_type)['impact']
        )

        significant_figures = rd.min_significant_figures(impact_factor.value)
        return impact_factor.value, significant_figures, impact_factor.min, impact_factor.max, ["End of life is not included in the calculation"]

    def impact_use(self, impact_type: str, duration: float) -> ComputedImpacts:
        raise NotImplementedError




import boaviztapi.utils.roundit as rd
from boaviztapi import config
from boaviztapi.model.boattribute import Boattribute
from boaviztapi.model.component.component import Component, ComputedImpacts
from boaviztapi.model.impact import ImpactFactor
from boaviztapi.service.archetype import get_component_archetype, get_arch_value
from boaviztapi.service.factor_provider import get_impact_factor


class ComponentHDD(Component):
    NAME = "HDD"

    __DISK_TYPE = 'hdd'

    def __init__(self, archetype=get_component_archetype(config["default_ssd"], "ssd"), **kwargs):
        super().__init__(archetype=archetype, **kwargs)

        self.capacity = Boattribute(
            unit="GB",
            default=get_arch_value(archetype, 'manufacturer', 'default'),
            min=get_arch_value(archetype, 'manufacturer', 'default'),
            max=get_arch_value(archetype, 'manufacturer', 'default')
        )

    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        impact = ImpactFactor(
            value=get_impact_factor(item='hdd', impact_type=impact_type)['impact'],
            min=get_impact_factor(item='hdd', impact_type=impact_type)['impact'],
            max=get_impact_factor(item='hdd', impact_type=impact_type)['impact']
        )

        significant_figures = rd.min_significant_figures(impact.value)
        return impact.value, significant_figures, impact.min, impact.max, ["End of life is not included in the calculation"]


from abc import ABC
from typing import List, Union

from boaviztapi import config
from boaviztapi.model import ComputedImpacts
from boaviztapi.model.component import Component, ComponentCPU, ComponentRAM, ComponentSSD, ComponentHDD, \
    ComponentPowerSupply, ComponentCase, ComponentMotherboard, ComponentAssembly
from boaviztapi.model.device.device import Device
from boaviztapi.model.impact import ImpactFactor
from boaviztapi.model.usage import ModelUsageServer, ModelUsageCloud
import boaviztapi.utils.roundit as rd
from boaviztapi.service.archetype import get_server_archetype, get_arch_component, get_cloud_instance_archetype
from boaviztapi.service.factor_provider import get_impact_factor


class DeviceServer(Device):
    NAME = "SERVER"
    def __init__(self, archetype=get_server_archetype(config["default_server"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)
        self._cpu = None
        self._ram_list = None
        self._disk_list = None
        self._power_supply = None
        self._case = None
        self._motherboard = None
        self._assembly = None
        self._usage = None

        for attr, val in kwargs.items():
            if val is not None:
                self.__setattr__(attr, val)

    @property
    def cpu(self) -> ComponentCPU:
        if self._cpu is None:
            self._cpu = ComponentCPU(archetype=get_arch_component(self.archetype,"CPU"))
        return self._cpu

    @cpu.setter
    def cpu(self, value: List[ComponentCPU]) -> None:
        self._cpu = value

    @property
    def ram(self) -> List[ComponentRAM]:
        if self._ram_list is None:
            self._ram_list = [ComponentRAM(archetype=get_arch_component(self.archetype,"RAM"))]
        return self._ram_list

    @ram.setter
    def ram(self, value: List[ComponentRAM]) -> None:
        self._ram_list = value

    @property
    def disk(self) -> List[Union[ComponentSSD, ComponentHDD]]:
        if self._disk_list is None:
            self._disk_list = [ComponentSSD(archetype=get_arch_component(self.archetype,"SSD"))]
        return self._disk_list

    @disk.setter
    def disk(self, value: List[Union[ComponentSSD, ComponentHDD]]) -> None:
        self._disk_list = value

    @property
    def power_supply(self) -> ComponentPowerSupply:
        if self._power_supply is None:
            self._power_supply = ComponentPowerSupply(archetype=get_arch_component(self.archetype,"POWER_SUPPLY"))
        return self._power_supply

    @power_supply.setter
    def power_supply(self, value: List[ComponentPowerSupply]) -> None:
        self._power_supply = value

    @property
    def case(self) -> ComponentCase:
        if self._case is None:
            self._case = ComponentCase(archetype=get_arch_component(self.archetype,"CASE"))
        return self._case

    @case.setter
    def case(self, value: ComponentCase) -> None:
        self._case = value

    @property
    def case_type(self) -> str:
        return self.case.case_type

    @case_type.setter
    def case_type(self, value: str) -> None:
        self.case.case_type = value

    @property
    def motherboard(self) -> ComponentMotherboard:
        if self._motherboard is None:
            self._motherboard = ComponentMotherboard(archetype=get_arch_component(self.archetype,"MOTHERBOARD"))
        return self._motherboard

    @motherboard.setter
    def motherboard(self, value: ComponentMotherboard) -> None:
        self._motherboard = value

    @property
    def assembly(self) -> ComponentAssembly:
        if self._assembly is None:
            self._assembly = ComponentAssembly(archetype=get_arch_component(self.archetype,"ASSEMBLY"))
        return self._assembly

    @assembly.setter
    def assembly(self, value: ComponentAssembly) -> None:
        self._assembly = value

    @property
    def usage(self) -> ModelUsageServer:
        if self._usage is None:
            self._usage = ModelUsageServer(archetype=get_arch_component(self.archetype,"USAGE"))
        return self._usage

    @usage.setter
    def usage(self, value: ModelUsageServer) -> None:
        self._usage = value

    @property
    def components(self) -> List[Component]:
        return [self.assembly] + [self.cpu] + self.ram + self.disk + [self.power_supply] + [self.case] + [
            self.motherboard]

    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        impacts = []
        min_impacts = []
        max_impacts = []
        significant_figures = []
        warnings = []

        try:
            for component in self.components:
                impact, sign_fig, min_impact, max_impact, c_warning = getattr(component, f'impact_embedded')(impact_type)

                impacts.append(impact*component.units.value)
                min_impacts.append(min_impact*component.units.min)
                max_impacts.append(max_impact*component.units.max)

                significant_figures.append(sign_fig)
                warnings = warnings + c_warning
            return sum(impacts), min(significant_figures), sum(min_impacts), sum(max_impacts), warnings

        except NotImplementedError:
            impact = get_impact_factor(item='SERVER', impact_type=impact_type)["impact"]
            min_impacts = get_impact_factor(item='SERVER', impact_type=impact_type)["impact"]
            max_impacts = get_impact_factor(item='SERVER', impact_type=impact_type)["impact"]
            significant_figures = rd.significant_number(impact)

            warnings = ["Generic data used for impact calculation."]

            return impact, significant_figures, min_impacts, max_impacts, warnings

    def impact_use(self, impact_type: str, duration: float) -> ComputedImpacts:
        impact_factor = self.usage.elec_factors[impact_type]

        if not self.usage.avg_power.is_set():
            modeled_consumption = self.model_power_consumption()
            self.usage.avg_power.set_completed(
                modeled_consumption.value,
                min=modeled_consumption.min,
                max=modeled_consumption.max
            )

        impact = impact_factor.value * (self.usage.avg_power.value / 1000) * self.usage.use_time_ratio.value * duration
        min_impact = impact_factor.min * (self.usage.avg_power.min / 1000) * self.usage.use_time_ratio.min * duration
        max_impact = impact_factor.max * (self.usage.avg_power.max / 1000) * self.usage.use_time_ratio.max * duration

        sig_fig = self.__compute_significant_numbers(impact_factor.value)
        return impact, sig_fig, min_impact, max_impact, []

    def model_power_consumption(self):
        conso_cpu = ImpactFactor(
            value=self.cpu.model_power_consumption().value*self.cpu.units.value,
            min=self.cpu.model_power_consumption().min*self.cpu.units.min,
            max=self.cpu.model_power_consumption().max * self.cpu.units.max,
        )
        conso_ram = ImpactFactor(
            value=0,
            min=0,
            max=0
        )
        for ram_unit in self.ram:
            conso_ram.value = conso_ram.value + ram_unit.model_power_consumption().value * ram_unit.units.value
            conso_ram.min = conso_ram.min + ram_unit.model_power_consumption().min * ram_unit.units.min
            conso_ram.max = conso_ram.max + ram_unit.model_power_consumption().max * ram_unit.units.max

        return ImpactFactor(
                value=(conso_cpu.value + conso_ram.value) * (1 + self.usage.other_consumption_ratio.value),
                min=(conso_cpu.min + conso_ram.min) * (1 + self.usage.other_consumption_ratio.min),
                max=(conso_cpu.max + conso_ram.max) * (1 + self.usage.other_consumption_ratio.max)
        )

    def __compute_significant_numbers(self, impact_factor: float) -> int:
        return rd.min_significant_figures(self.usage.avg_power.value, impact_factor)


class DeviceCloudInstance(DeviceServer, ABC):

    def __init__(self, archetype=get_cloud_instance_archetype(config["default_cloud"], config["default_cloud_provider"]), **kwargs):
        super().__init__(**kwargs, archetype=archetype)

    @property
    def usage(self) -> ModelUsageCloud:
        if self._usage is None:
            self._usage = ModelUsageCloud(archetype=get_arch_component(self.archetype,"USAGE"))
        return self._usage

    @usage.setter
    def usage(self, value: ModelUsageCloud) -> None:
        self._usage = value

    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        impact, sign_fig, min_impact, max_impact, c_warning = super().impact_embedded(impact_type)
        return (impact / self.usage.instance_per_server.value), sign_fig, min_impact, max_impact, c_warning

    def impact_use(self, impact_type: str, duration: int) -> ComputedImpacts:
        impact, sign_fig, min_impact, max_impact, c_warning = super().impact_use(impact_type, duration)
        return (impact / self.usage.instance_per_server.value), sign_fig, min_impact, max_impact, c_warning


from typing import Tuple

import boaviztapi.utils.roundit as rd
from boaviztapi import config
from boaviztapi.model import ComputedImpacts
from boaviztapi.model.boattribute import Boattribute
from boaviztapi.model.component.component import Component
from boaviztapi.model.impact import ImpactFactor
from boaviztapi.service.archetype import get_arch_value, get_component_archetype
from boaviztapi.service.factor_provider import get_impact_factor


class ComponentCase(Component):
    AVAILABLE_CASE_TYPE = ['blade', 'rack']
    NAME = "CASE"

    def __init__(self, archetype=get_component_archetype(config["default_case"], "case"), **kwargs):
        super().__init__(archetype=archetype, **kwargs)
        self.case_type = Boattribute(
            default=get_arch_value(archetype, 'case_type', 'default'),
            min=get_arch_value(archetype, 'case_type', 'min'),
            max=get_arch_value(archetype, 'case_type', 'max')
        )
    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        if self.case_type.value == 'rack':
            return self.__impact_manufacture_rack(impact_type)
        elif self.case_type.value == 'blade':
            return self.__impact_manufacture_blade(impact_type)

    def __impact_manufacture_rack(self, impact_type: str) -> ComputedImpacts:
        impact_factor = ImpactFactor(
            value=get_impact_factor(item='case', impact_type=impact_type)['rack']['impact'],
            min=get_impact_factor(item='case', impact_type=impact_type)['rack']['impact'],
            max=get_impact_factor(item='case', impact_type=impact_type)['rack']['impact']
        )

        significant_figures = rd.min_significant_figures(impact_factor.value)

        if self.case_type.is_archetype() and self.case_type.value == 'rack':
            blade_impact = self.__impact_manufacture_blade(impact_type)
            if blade_impact[0] > impact_factor.value:
                return impact_factor.value, significant_figures, impact_factor.min, blade_impact[3], ["End of life is not included in the calculation"]
            else:
                return impact_factor.value, significant_figures, blade_impact[2], impact_factor.max, ["End of life is not included in the calculation"]
        return impact_factor.value, significant_figures, impact_factor.min, impact_factor.max, ["End of life is not included in the calculation"]

    def __impact_manufacture_blade(self, impact_type: str) -> ComputedImpacts:
        impact_blade_server, impact_blade_16_slots = self.__get_impact_constants_blade(impact_type)

        impact = self.__compute_impact_manufacture_blade(impact_blade_server, impact_blade_16_slots)

        significant_figures = self.__compute_significant_numbers(impact_blade_server.value, impact_blade_16_slots.value)

        if self.case_type.is_archetype() and self.case_type.value == 'blade':
            rack_impact = self.__impact_manufacture_rack(impact_type)
            if rack_impact[0] > impact.value:
                return impact.value, significant_figures, impact.min, rack_impact[3], ["End of life is not included in the calculation"]
            else:
                return impact.value, significant_figures, rack_impact[2], impact.max, ["End of life is not included in the calculation"]

        return impact.value, significant_figures, impact.min, impact.max, ["End of life is not included in the calculation"]

    def __get_impact_constants_blade(self, impact_type: str) -> Tuple[ImpactFactor, ImpactFactor]:
        impact_blade_server = ImpactFactor(
            value=get_impact_factor(item='case', impact_type=impact_type)['blade']['impact_blade_server'],
            min=get_impact_factor(item='case', impact_type=impact_type)['blade']['impact_blade_server'],
            max=get_impact_factor(item='case', impact_type=impact_type)['blade']['impact_blade_server'],
        )
        impact_blade_16_slots = ImpactFactor(
            value=get_impact_factor(item='case', impact_type=impact_type)['blade']['impact_blade_16_slots'],
            min=get_impact_factor(item='case', impact_type=impact_type)['blade']['impact_blade_16_slots'],
            max=get_impact_factor(item='case', impact_type=impact_type)['blade']['impact_blade_16_slots'],
        )

        return impact_blade_server, impact_blade_16_slots

    @staticmethod
    def __compute_impact_manufacture_blade(impact_blade_server: ImpactFactor, impact_blade_16_slots: ImpactFactor) -> ImpactFactor:
        return ImpactFactor(
            value=(impact_blade_16_slots.value / 16) + impact_blade_server.value,
            min=(impact_blade_16_slots.min / 16) + impact_blade_server.min,
            max=(impact_blade_16_slots.max / 16) + impact_blade_server.max
        )

    @staticmethod
    def __compute_significant_numbers(impact_blade_server: float, impact_blade_16_slots: float) -> int:
        return rd.min_significant_figures(impact_blade_server, impact_blade_16_slots)

    def impact_use(self, impact_type: str, duration: float) -> ComputedImpacts:
        raise NotImplementedError


import os

from boaviztapi import config, data_dir
from boaviztapi.model.boattribute import Boattribute
from boaviztapi.service.archetype import get_arch_value, get_server_archetype, get_cloud_instance_archetype
from boaviztapi.service.factor_provider import get_available_countries, get_electrical_impact_factor, \
    get_electrical_min_max

_cpu_profile_path = os.path.join(data_dir, 'consumption_profile/cpu/cpu_profile.csv')
_cloud_profile_path = os.path.join(data_dir, 'consumption_profile/cloud/cpu_profile.csv')
_server_profile_path = os.path.join(data_dir, 'consumption_profile/server/server_profile.csv')

class ModelUsage:

    _DAYS_IN_HOURS = 24
    _YEARS_IN_HOURS = 24 * 365

    def __init__(self, archetype, **kwargs):
        self.archetype = archetype
        self.avg_power = Boattribute(
            unit="W",
            default=get_arch_value(archetype, 'avg_power', 'default'),
            min=get_arch_value(archetype, 'avg_power', 'min'),
            max=get_arch_value(archetype, 'avg_power', 'max')
        )
        self.time_workload = Boattribute(
            unit="%",
            default=get_arch_value(archetype, 'time_workload', 'default'),
            min=get_arch_value(archetype, 'time_workload', 'min'),
            max=get_arch_value(archetype, 'time_workload', 'max')
        )
        self.consumption_profile = None
        self.usage_location = Boattribute(
            unit="CodSP3 - NCS Country Codes - NATO",
            default=get_arch_value(archetype, 'usage_location', 'default'),
            min=get_arch_value(archetype, 'usage_location', 'min'),
            max=get_arch_value(archetype, 'usage_location', 'max')
        )
        self.use_time_ratio = Boattribute(
            unit="/1",
            default=get_arch_value(archetype, 'use_time_ratio', 'default'),
            min=get_arch_value(archetype, 'use_time_ratio', 'min'),
            max=get_arch_value(archetype, 'use_time_ratio', 'max')
        )
        self.hours_life_time = Boattribute(
            unit="hours",
            default=get_arch_value(archetype, 'hours_life_time', 'default'),
            min=get_arch_value(archetype, 'hours_life_time', 'min'),
            max=get_arch_value(archetype, 'hours_life_time', 'max')
        )
        self.elec_factors = {
            "gwp": Boattribute(unit="kg CO2eq/kWh", complete_function=self._complete_gwp),
            "adp": Boattribute(unit="kg Sbeq/kWh", complete_function=self._complete_adp),
            "pe": Boattribute(unit="MJ/kWh", complete_function=self._complete_pe),
            "gwppb": Boattribute(unit="kg CO2eq/kWh", complete_function=self._complete_gwppb),
            "gwppf": Boattribute(unit="kg CO2eq/kWh", complete_function=self._complete_gwppf),
            "gwpplu": Boattribute(unit="kg CO2eq/kWh", complete_function=self._complete_gwpplu),
            "ir": Boattribute(unit="kg U235eq/kWh", complete_function=self._complete_ir),
            "lu": Boattribute(unit="No dimension/kWh", complete_function=self._complete_lu),
            "odp": Boattribute(unit="kg CFC-11eq/kWh", complete_function=self._complete_odp),
            "pm": Boattribute(unit="Disease occurrence/kWh", complete_function=self._complete_pm),
            "pocp": Boattribute(unit="kg NMVOCeq/kWh", complete_function=self._complete_pocp),
            "wu": Boattribute(unit="m3eq/kWh", complete_function=self._complete_wu),
            "mips": Boattribute(unit="kg/kWh", complete_function=self._complete_mips),
            "adpe": Boattribute(unit="kg Sbeq/kWh", complete_function=self._complete_adpe),
            "adpf": Boattribute(unit="MJ/kWh", complete_function=self._complete_adpf),
            "ap": Boattribute(unit="mol H+eq/kWh", complete_function=self._complete_ap),
            "ctue": Boattribute(unit="CTUe/kWh", complete_function=self._complete_ctue),
            "ctuh_c": Boattribute(unit="CTUh/kWh", complete_function=self._complete_ctuh_c),
            "ctuh_nc": Boattribute(unit="CTUh/kWh", complete_function=self._complete_ctuh_nc),
            "epf": Boattribute(unit="kg Peq/kWh", complete_function=self._complete_epf),
            "epm": Boattribute(unit="kg Neq/kWh", complete_function=self._complete_epm),
            "ept": Boattribute(unit="mol Neq/kWh", complete_function=self._complete_ept),
        }

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def _complete_impact_factor(self, impact_criteria, impact_criteria_proxy):
        if impact_criteria is None:
            raise NotImplementedError

        if not self.usage_location.has_value():
            self.usage_location.set_default(config["default_location"])

        if self.usage_location.value not in get_available_countries(reverse=True):
            raise NotImplementedError

        factor = get_electrical_impact_factor(self.usage_location.value, impact_criteria_proxy)

        if self.usage_location.is_default():
            self.elec_factors.get(impact_criteria).set_default(factor["value"], source=str(factor["source"]))
            self.elec_factors.get(impact_criteria).min = float(get_electrical_min_max(impact_criteria_proxy, "min"))
            self.elec_factors.get(impact_criteria).max =  float(get_electrical_min_max(impact_criteria_proxy, "max"))
        else:
            self.elec_factors.get(impact_criteria).set_completed(factor["value"],
                                                             source=str(factor["source"]),
                                                             min=factor["value"],
                                                             max=factor["value"])

    def _complete_gwp(self):
        self._complete_impact_factor("gwp", "gwp")
    def _complete_adp(self):
        self._complete_impact_factor("adp", "adpe")
    def _complete_pe(self):
        self._complete_impact_factor("pe", "pe")
    def _complete_gwppb(self):
        self._complete_impact_factor("gwppb", "gwppb")
    def _complete_gwppf(self):
        self._complete_impact_factor("gwppf", "gwppf")
    def _complete_gwpplu(self):
        self._complete_impact_factor("gwpplu", "gwpplu")
    def _complete_ir(self):
        self._complete_impact_factor("ir", "ir")
    def _complete_lu(self):
        self._complete_impact_factor("lu", "lu")
    def _complete_odp(self):
        self._complete_impact_factor("odp", "odp")
    def _complete_pm(self):
        self._complete_impact_factor("pm", "pm")
    def _complete_pocp(self):
        self._complete_impact_factor("pocp", "pocp")
    def _complete_wu(self):
        self._complete_impact_factor("wu", "wu")
    def _complete_mips(self):
        self._complete_impact_factor("mips", "mips")
    def _complete_adpe(self):
        self._complete_impact_factor("adpe", "adpe")
    def _complete_adpf(self):
        self._complete_impact_factor("adpf", "adpf")
    def _complete_ap(self):
        self._complete_impact_factor("ap", "ap")
    def _complete_ctue(self):
        self._complete_impact_factor("ctue", "ctue")
    def _complete_ctuh_c(self):
        self._complete_impact_factor("ctuh_c", "ctuh_c")
    def _complete_ctuh_nc(self):
        self._complete_impact_factor("ctuh_nc", "ctuh_nc")
    def _complete_epf(self):
        self._complete_impact_factor("epf", "epf")
    def _complete_epm(self):
        self._complete_impact_factor("epm", "epm")
    def _complete_ept(self):
        self._complete_impact_factor("ept", "ept")

class ModelUsageServer(ModelUsage):

    def __init__(self, archetype=get_server_archetype(config["default_server"]).get("USAGE"), **kwargs):
        super().__init__(archetype=archetype, **kwargs)

        self.other_consumption_ratio = Boattribute(
            unit="ratio /1",
            default=get_arch_value(archetype, 'other_consumption_ratio', 'default'),
            min=get_arch_value(archetype, 'other_consumption_ratio', 'min'),
            max=get_arch_value(archetype, 'other_consumption_ratio', 'max')
        )

class ModelUsageCloud(ModelUsageServer):
    def __init__(self, archetype=get_cloud_instance_archetype(config["default_cloud"], config["default_cloud_provider"]).get("USAGE"), **kwargs):
        super().__init__(archetype=archetype, **kwargs)
        self.instance_per_server = Boattribute(
            default=get_arch_value(archetype, 'instance_per_server', 'default'),
            min=get_arch_value(archetype, 'instance_per_server', 'min'),
            max=get_arch_value(archetype, 'instance_per_server', 'max')
        )


import os
from typing import Tuple

import pandas as pd

import boaviztapi.utils.roundit as rd
from boaviztapi import config, data_dir
from boaviztapi.model import ComputedImpacts
from boaviztapi.model.boattribute import Boattribute
from boaviztapi.model.component.component import Component
from boaviztapi.model.consumption_profile import CPUConsumptionProfileModel
from boaviztapi.model.impact import ImpactFactor
from boaviztapi.service.archetype import get_component_archetype, get_arch_value
from boaviztapi.service.factor_provider import get_impact_factor
from boaviztapi.utils.fuzzymatch import fuzzymatch_attr_from_pdf, fuzzymatch_attr_from_cpu_name

_cpu_specs = pd.read_csv(os.path.join(data_dir, 'crowdsourcing/cpu_specs.csv'))
_family_df = pd.read_csv(os.path.join(data_dir, 'crowdsourcing/cpu_manufacture.csv'))


def attributes_from_cpu_name(cpu_name: str):
    return fuzzymatch_attr_from_cpu_name(cpu_name, _cpu_specs)


class ComponentCPU(Component):
    NAME = "CPU"
    name_completion = False

    def __init__(self, archetype=get_component_archetype(config["default_cpu"], "cpu"), **kwargs):
        super().__init__(archetype=archetype, **kwargs)
        self.core_units = Boattribute(
            complete_function=self._complete_from_name,
            default=get_arch_value(archetype, 'core_units', 'default'),
            min=get_arch_value(archetype, 'core_units', 'min'),
            max=get_arch_value(archetype, 'core_units', 'max')
        )
        self.die_size_per_core = Boattribute(
            complete_function=self._complete_die_size_per_core,
            unit="cm2",
            default=get_arch_value(archetype, 'die_size_per_core', 'default'),
            min=get_arch_value(archetype, 'die_size_per_core', 'min'),
            max=get_arch_value(archetype, 'die_size_per_core', 'max')
        )
        self.die_size = Boattribute(
            complete_function=self._complete_from_name,
            unit="cm2",
            default=get_arch_value(archetype, 'die_size', 'default'),
            min=get_arch_value(archetype, 'die_size', 'min'),
            max=get_arch_value(archetype, 'die_size', 'max')
        )
        self.model_range = Boattribute(
            default=get_arch_value(archetype, 'model_range', 'default'),
            min=get_arch_value(archetype, 'model_range', 'min'),
            max=get_arch_value(archetype, 'model_range', 'max')
        )
        self.manufacturer = Boattribute(
            default=get_arch_value(archetype, 'manufacturer', 'default'),
            min=get_arch_value(archetype, 'manufacturer', 'min'),
            max=get_arch_value(archetype, 'manufacturer', 'max')
        )
        self.family = Boattribute(
            complete_function=self._complete_from_name,
            default=get_arch_value(archetype, 'family', 'default'),
            min=get_arch_value(archetype, 'family', 'min'),
            max=get_arch_value(archetype, 'family', 'max')
        )
        self.name = Boattribute(
            default=get_arch_value(archetype, 'name', 'default'),
            min=get_arch_value(archetype, 'name', 'min'),
            max=get_arch_value(archetype, 'name', 'max')
        )
        self.tdp = Boattribute(
            complete_function=self._complete_from_name,
            unit="W",
            default=get_arch_value(archetype, 'tdp', 'default'),
            min=get_arch_value(archetype, 'tdp', 'min'),
            max=get_arch_value(archetype, 'tdp', 'max')
        )
        self.usage.avg_power.add_warning("value for one cpu unit")

    # IMPACT COMPUTATION
    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        core_impact, cpu_die_impact, cpu_impact = self.__get_impact_constants(impact_type)
        sign_figures = self.__compute_significant_numbers(core_impact.value, cpu_die_impact.value, cpu_impact.value)
        impact = self.__compute_impact_manufacture(core_impact, cpu_die_impact, cpu_impact)

        return impact.value, sign_figures, impact.min, impact.max, ["End of life is not included in the calculation"]

    def impact_use(self, impact_type: str, duration: float) -> ComputedImpacts:
        impact_factor = self.usage.elec_factors[impact_type]

        if not self.usage.avg_power.is_set():
            modeled_consumption = self.model_power_consumption()
            self.usage.avg_power.set_completed(
                modeled_consumption.value,
                min=modeled_consumption.min,
                max=modeled_consumption.max
            )

        impact = ImpactFactor(
            value=impact_factor.value * (
                    self.usage.avg_power.value / 1000) * self.usage.use_time_ratio.value * duration,
            min=impact_factor.min * (
                    self.usage.avg_power.min / 1000) * self.usage.use_time_ratio.min * duration,
            max=impact_factor.max * (
                    self.usage.avg_power.max / 1000) * self.usage.use_time_ratio.max * duration
        )

        sig_fig = self.__compute_significant_numbers_usage(impact_factor.value)
        return impact.value, sig_fig, impact.min, impact.max, []

    def __compute_significant_numbers_usage(self, impact_factor: float) -> int:
        return rd.min_significant_figures(self.usage.avg_power.value, self.usage.use_time_ratio.value, impact_factor)

    # TODO: compute min & max
    def model_power_consumption(self) -> ImpactFactor:
        self.usage.consumption_profile = CPUConsumptionProfileModel()

        self.usage.consumption_profile.compute_consumption_profile_model(cpu_manufacturer=self.manufacturer.value,
                                                                         cpu_model_range=self.model_range.value,
                                                                         cpu_tdp=self.tdp.value)
        if type(self.usage.time_workload.value) in (float, int):
            self.usage.avg_power.set_completed(self.usage.consumption_profile.apply_consumption_profile(self.usage.time_workload.value))
        else:
            self.usage.avg_power.set_completed(self.usage.consumption_profile.apply_multiple_workloads(self.usage.time_workload.value))

        return ImpactFactor(
                value=rd.round_to_sigfig(self.usage.avg_power.value, 5),
                min=rd.round_to_sigfig(self.usage.avg_power.value, 5),
                max=rd.round_to_sigfig(self.usage.avg_power.value, 5)
        )

    def __get_impact_constants(self, impact_type: str) -> Tuple[ImpactFactor, ImpactFactor, ImpactFactor]:
        core_impact = ImpactFactor(
            value=get_impact_factor(item='cpu', impact_type=impact_type)['constant_core_impact'],
            min=get_impact_factor(item='cpu', impact_type=impact_type)['constant_core_impact'],
            max=get_impact_factor(item='cpu', impact_type=impact_type)['constant_core_impact']
        )
        cpu_die_impact = ImpactFactor(
            value=get_impact_factor(item='cpu', impact_type=impact_type)['die_impact'],
            min=get_impact_factor(item='cpu', impact_type=impact_type)['die_impact'],
            max=get_impact_factor(item='cpu', impact_type=impact_type)['die_impact']
        )
        cpu_impact = ImpactFactor(
            value=get_impact_factor(item='cpu', impact_type=impact_type)['impact'],
            min=get_impact_factor(item='cpu', impact_type=impact_type)['impact'],
            max=get_impact_factor(item='cpu', impact_type=impact_type)['impact']
        )

        return core_impact, cpu_die_impact, cpu_impact

    def __compute_significant_numbers(self, core_impact: float, cpu_die_impact: float, cpu_impact: float) -> int:
        return rd.min_significant_figures(self.die_size_per_core.value, core_impact, cpu_die_impact, cpu_impact)

    def __compute_impact_manufacture(self, core_impact: ImpactFactor, cpu_die_impact: ImpactFactor, cpu_impact: ImpactFactor) -> ImpactFactor:
        return ImpactFactor(
            value=(self.core_units.value * self.die_size_per_core.value + core_impact.value) * cpu_die_impact.value + cpu_impact.value,
            min=(self.core_units.min * self.die_size_per_core.min + core_impact.min) * cpu_die_impact.min + cpu_impact.min,
            max=(self.core_units.max * self.die_size_per_core.max + core_impact.max) * cpu_die_impact.max + cpu_impact.max
        )

    # COMPLETION
    def _complete_die_size_per_core(self):
        self._complete_from_name()
        corrected_family = None
        # If we can have a die_size and core_units, we can compute the die_size_per_core
        if self.die_size.has_value() and self.core_units.has_value():
            self.die_size_per_core.set_completed(value=float(self.die_size.value/self.core_units.value), source="die_size/core_units")
            return None

        sub = _family_df.copy()
        # If we don't have a family, we use the die_size_per_core from the archetype
        if self.die_size_per_core.has_value() and not self.family.has_value():
            return None

        if self.family.has_value():
            # Check if the family matches one of the families in the database and correct it if necessary
            corrected_family = fuzzymatch_attr_from_pdf(self.family.value, "family", sub)
            if corrected_family != self.family.value and corrected_family is not None:
                self.family.set_changed(corrected_family)

            tmp = sub[sub['family'] == corrected_family]

            if len(tmp) > 0:
                sub = tmp.copy()

            if self.core_units.has_value() and corrected_family is not None:
                # Find the closest line to the number of cores provided by the user
                sub['core_dif'] = sub[['core_units']].apply(lambda x: abs(x[0] - self.core_units.value), axis=1)
                sub = sub.sort_values(by='core_dif', ascending=True)
                row = sub.iloc[0]

                # If we have only one row but the number of cores is different, we use the max and min values of the dataframe
                if row['core_dif'] != 0 and len(sub) == 1:
                    self.die_size_per_core.set_completed(float(row['die_size_per_core']), source=row['Source'])
                    self.die_size_per_core.min = float(float(_family_df['die_size_per_core'].min()))
                    self.die_size_per_core.max = float(float(_family_df['die_size_per_core'].max()))
                    return

                row2 = sub.iloc[1]
                self.die_size_per_core.set_completed(float(row['die_size_per_core']), source=row['Source'])

                if row['core_dif'] == 0:
                    self.die_size_per_core.min = float(row['die_size_per_core'])
                    self.die_size_per_core.max = float(row['die_size_per_core'])

                elif float(row2['die_size_per_core']) > float(row['die_size_per_core']):
                    self.die_size_per_core.min = float(row['die_size_per_core'])
                    self.die_size_per_core.max = float(row2['die_size_per_core'])

                else:
                    self.die_size_per_core.min = float(row2['die_size_per_core'])
                    self.die_size_per_core.max = float(row['die_size_per_core'])

                return

        source_family = self.family.value if corrected_family is not None else "all families"

        # If we don't have a number of cores, we use the average die size per core for a given family (if provided)
        self.die_size_per_core.set_completed(float(sub['die_size_per_core'].mean()), source=f"Average for {source_family}")
        self.die_size_per_core.min = float(float(sub['die_size_per_core'].min()))
        self.die_size_per_core.max = float(float(sub['die_size_per_core'].max()))

    def _complete_from_name(self):
        if self.name.has_value() and not self.name_completion:
            compute_min_max = False
            if self.name.min != self.name.value or self.name.max != self.name.value:
                compute_min_max = True

            name, manufacturer, family, model_range, tdp, cores, die_size, die_size_source, source  = attributes_from_cpu_name(self.name.value)

            if compute_min_max:
                name_min, manufacturer_min, family_min, model_range_min, tdp_min, cores_min, die_size_min, die_size_source_min, source_min  = attributes_from_cpu_name(self.name.min)
                name_max, manufacturer_max, family_max, model_range_max, tdp_max, cores_max, die_size_max, die_size_source_max, source_max  = attributes_from_cpu_name(self.name.max)
            else:
                name_min, manufacturer_min, family_min, model_range_min, tdp_min, cores_min, die_size_min, die_size_source_min, source_min  = name, manufacturer, family, model_range, tdp, cores, die_size, die_size_source, source
                name_max, manufacturer_max, family_max, model_range_max, tdp_max, cores_max, die_size_max, die_size_source_max, source_max  = name, manufacturer, family, model_range, tdp, cores, die_size, die_size_source, source

            if name is not None:
                self.name.set_completed(name, min=name_min, max=name_max, source="fuzzy match")
            if manufacturer is not None:
                self.manufacturer.set_completed(manufacturer, min=manufacturer_min, max=manufacturer_max, source=f"Completed from name name based on {source}.")
            if family is not None:
                self.family.set_completed(family, min=family_min, max=family_max, source=f"Completed from name name based on {source}.")
            if model_range is not None:
                self.model_range.set_completed(model_range, min=model_range_min, max=model_range_max, source=f"Completed from name name based on {source}.")
            if tdp is not None:
                self.tdp.set_completed(tdp, min=tdp_min, max=tdp_max, source=f"Completed from name name based on {source}.")
            if cores is not None:
                self.core_units.set_completed(cores, min=cores_min, max=cores_max, source=f"Completed from name name based on {source}.")
            if die_size is not None:
                # divide by 100 to convert mm2 into cm2
                self.die_size.set_completed(die_size/100, min=die_size_min/100, max=die_size_max/100, source=f"{die_size_source} : Completed from name name based on {source}.")
            self.name_completion = True



from typing import Optional

from boaviztapi.dto.base_dto import BaseDTO
from boaviztapi.dto.usage import Usage


class ComponentDTO(BaseDTO):
    units: Optional[int]
    usage: Optional[Usage] = Usage()



import json
import subprocess

import markdown
import toml
from fastapi.middleware.cors import CORSMiddleware
import os
from fastapi import FastAPI
from fastapi.openapi.utils import get_openapi
from mangum import Mangum

from boaviztapi.routers.component_router import component_router
from boaviztapi.routers.consumption_profile_router import consumption_profile
from boaviztapi.routers.peripheral_router import peripheral_router
from boaviztapi.routers.server_router import server_router
from boaviztapi.routers.cloud_router import cloud_router
from boaviztapi.routers.terminal_router import terminal_router
from boaviztapi.routers.utils_router import utils_router

from fastapi.responses import HTMLResponse

# Serverless frameworks adds a 'stage' prefix to the route used to serve applications
# We have to manage it to expose openapi doc on aws and generate proper links.
stage = os.environ.get('STAGE', None)
openapi_prefix = f"/{stage}" if stage else "/"
app = FastAPI(root_path=openapi_prefix)  # Here is the magic
version = toml.loads(open(os.path.join(os.path.dirname(__file__), '../pyproject.toml'), 'r').read())['tool']['poetry']['version']


origins = json.loads(os.getenv("ALLOWED_ORIGINS", '["*"]'))

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(server_router)
app.include_router(cloud_router)
app.include_router(terminal_router)
app.include_router(peripheral_router)
app.include_router(component_router)
app.include_router(utils_router)
app.include_router(consumption_profile)

if __name__ == '__main__':
    import uvicorn

    uvicorn.run('main:app', host='localhost', port=5000, reload=True, debug=True)


@app.on_event("startup")
def my_schema():
    intro = open(os.path.join(os.path.dirname(__file__), 'routers/openapi_doc/intro_openapi.md'), 'r', encoding='utf-8')
    openapi_schema = get_openapi(
        title="BOAVIZTAPI - DEMO",
        version=version,
        description=markdown.markdown(intro.read()),
        routes=app.routes,
        servers=app.servers,
    )
    app.openapi_schema = openapi_schema

    return app.openapi_schema


# Wrapper for aws/lambda serverless app
handler = Mangum(app)


@app.get("/", response_class=HTMLResponse)
async def welcome_page():
    html_content = """
    <html>
        <head>
            <title>BOAVIZTAPI</title>
            <style>
                * {
                    font-family: sans-serif;
                }
            </style>
        </head>
        <body>
            <p align="center">
                <img src="https://boavizta.org/media/site/d84925bc94-1642413712/boavizta-logo-4.png" width="100">
            </p>
            <h1 align="center">
              Welcome to BOAVIZTAPI
            </h1>
            <h2 align="center">
              Multicriteria & multistep impacts evaluations for digital assets
            </h2>
            </br>
            <h3 align="center">See our github repository : <a href="https://github.com/Boavizta/boaviztapi">LINK</a></h2>
            <h3 align="center">See OpenAPI specs (swagger) : <a href="docs">LINK</a></h2>
            <h3 align="center">See our complete documentation : <a href="https://doc.api.boavizta.org/">LINK</a></h2>
            <h3 align="center">See the other resources of Boavizta : <a href="https://boavizta.org/">LINK</a> </h2>
            %s
        </body>
    </html>
    """ % os.getenv('SPECIAL_MESSAGE', '')
    return HTMLResponse(content=html_content, status_code=200)



import pandas as pd
from pandas.core.series import Series
from rapidfuzz import process, fuzz
from typing import Tuple


def fuzzymatch_attr_from_cpu_name(cpu_name: str, df: pd.DataFrame) -> Tuple[str, str, str, str, int, int, int, str, str]:
    cpu_name = cpu_name.lower()
    score = df["name"].str.lower().apply(lambda x: fuzz.token_set_ratio(x, cpu_name))
    best = df.iloc[score.idxmax()]
    best = best.mask(best.isnull(), None)  # replace all NaN by None
    safe_int = lambda x: x if x is None else int(x)  # float to int but keep None
    return (
        best["name"],  # .name is reserved by pandas for indexes
        best.manufacturer,
        best.code_name,
        best.model_range,
        safe_int(best.tdp),
        safe_int(best.cores),
        safe_int(best.total_die_size),
        best.total_die_size_source,
        best.source,
    )


def fuzzymatch_attr_from_pdf(name: str, attr: str, pdf: pd.DataFrame) -> str:
    name_list = list(pdf[attr].unique())

    result = process.extractOne(name, name_list, scorer=fuzz.WRatio)
    if result is not None:
        result = result[0] if result[1] > 79.0 else None
    return result or None


def fuzzymatch(s: pd.Series, value: str, threshold: float = 90.0) -> pd.Series:
    return s.apply(lambda x: fuzz.token_sort_ratio(x.lower(), value.lower()) >= threshold)


def pandas() -> None:
    Series.fuzzymatch = fuzzymatch
    
    



import os
from typing import Optional

import pandas as pd

from boaviztapi import config
from boaviztapi.dto.component import ComponentDTO
from boaviztapi.dto.usage.usage import mapper_usage, Usage
from boaviztapi.model.component import ComponentCPU
from boaviztapi.service.archetype import get_component_archetype


_cpu_df = pd.read_csv(os.path.join(os.path.dirname(__file__), '../../data/crowdsourcing/cpu_manufacture.csv'))


class CPU(ComponentDTO):
    core_units: Optional[int] = None
    die_size: Optional[float] = None
    die_size_per_core: Optional[float] = None
    manufacturer: Optional[str] = None
    model_range: Optional[str] = None
    family: Optional[str] = None
    name: Optional[str] = None
    tdp: Optional[int] = None

def mapper_cpu(cpu_dto: CPU, archetype=get_component_archetype(config["default_cpu"], "cpu")) -> ComponentCPU:
    cpu_component = ComponentCPU(archetype=archetype)
    cpu_component.usage = mapper_usage(cpu_dto.usage or Usage(), archetype=archetype.get("USAGE"))

    if cpu_dto.units is not None:
        cpu_component.units.set_input(cpu_dto.units)

    if cpu_dto.family is not None:
        cpu_component.family.set_input(cpu_dto.family)

    if cpu_dto.name is not None:
        cpu_component.name.set_input(cpu_dto.name)

    if cpu_dto.core_units is not None:
        cpu_component.core_units.set_input(cpu_dto.core_units)

    if cpu_dto.tdp is not None:
        cpu_component.tdp.set_input(cpu_dto.tdp)

    if cpu_dto.model_range is not None:
        cpu_component.model_range.set_input(cpu_dto.model_range)

    if cpu_dto.die_size_per_core is not None:
        cpu_component.die_size_per_core.set_input(cpu_dto.die_size_per_core)

    if cpu_dto.die_size_per_core is not None:
        cpu_component.die_size_per_core.set_input(cpu_dto.die_size_per_core)

    elif cpu_dto.die_size is not None and cpu_dto.core_units is not None:
        die_size_per_core = cpu_dto.die_size / cpu_dto.core_units
        cpu_component.die_size_per_core.set_completed(
            die_size_per_core,
            source="INPUT : die_size / core_units",
            min=die_size_per_core,
            max=die_size_per_core
        )

    return cpu_component



from abc import abstractmethod, ABC

from boaviztapi import config
from boaviztapi.model import ComputedImpacts
from boaviztapi.model.boattribute import Boattribute
from boaviztapi.model.device import Device
import boaviztapi.utils.roundit as rd
from boaviztapi.model.usage import ModelUsage
from boaviztapi.service.archetype import get_arch_value, get_user_terminal_archetype, get_arch_component
from boaviztapi.service.factor_provider import get_impact_factor

class EndUserDevice(Device):
    NAME = None
    def __init__(self, archetype=None, **kwargs):
        super().__init__(archetype=archetype, **kwargs)

    @property
    def usage(self) -> ModelUsage:
        if self._usage is None:
            self._usage = ModelUsage(archetype=get_arch_component(self.archetype,"USAGE"))
        return self._usage

    @usage.setter
    def usage(self, value: int) -> None:
        self._usage = value
    @abstractmethod
    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        impact = float(get_impact_factor(item=self.NAME, impact_type=impact_type)["impact"])
        min_impacts = float(get_impact_factor(item=self.NAME, impact_type=impact_type)["impact"])
        max_impacts = float(get_impact_factor(item=self.NAME, impact_type=impact_type)["impact"])
        significant_figures = rd.significant_number(impact)

        warnings = ["Generic data used for impact calculation."]

        return impact, significant_figures, min_impacts, max_impacts, warnings

    @abstractmethod
    def impact_use(self, impact_type: str, duration: float) -> ComputedImpacts:
        impact_factor = self.usage.elec_factors[impact_type]
        impact = impact_factor.value * (self.usage.avg_power.value / 1000) * self.usage.use_time_ratio.value * duration
        min_impact = impact_factor.min * (self.usage.avg_power.min / 1000) * self.usage.use_time_ratio.min * duration
        max_impact = impact_factor.max * (self.usage.avg_power.max / 1000) * self.usage.use_time_ratio.max * duration

        sig_fig = rd.min_significant_figures(self.usage.avg_power.value, self.usage.use_time_ratio.value, impact_factor.value)
        return impact, sig_fig, min_impact, max_impact, []

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

class DeviceLaptop(EndUserDevice, ABC):
    NAME = "LAPTOP"
    def __init__(self, archetype=get_user_terminal_archetype(config["default_laptop"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)
        self.type = Boattribute(
            default=get_arch_value(archetype, 'type', 'default'),
            min=get_arch_value(archetype, 'type', 'min'),
            max=get_arch_value(archetype, 'type', 'max')
        )
    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        impact = float(get_impact_factor(item=self.NAME, impact_type=impact_type)[self.type.value]["impact"])
        min_impacts = float(get_impact_factor(item=self.NAME, impact_type=impact_type)[self.type.value]["impact"])
        max_impacts = float(get_impact_factor(item=self.NAME, impact_type=impact_type)[self.type.value]["impact"])
        significant_figures = rd.significant_number(impact)

        warnings = ["Generic data used for impact calculation."]

        return impact, significant_figures, min_impacts, max_impacts, warnings

class DeviceDesktop(EndUserDevice, ABC):
    NAME = "DESKTOP"
    def __init__(self, archetype=get_user_terminal_archetype(config["default_desktop"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)
        self.type = Boattribute(
            default=get_arch_value(archetype, 'type', 'default'),
            min=get_arch_value(archetype, 'type', 'min'),
            max=get_arch_value(archetype, 'type', 'max')
        )

    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        impact = float(get_impact_factor(item=self.NAME, impact_type=impact_type)[self.type.value]["impact"])
        min_impacts = float(get_impact_factor(item=self.NAME, impact_type=impact_type)[self.type.value]["impact"])
        max_impacts = float(get_impact_factor(item=self.NAME, impact_type=impact_type)[self.type.value]["impact"])
        significant_figures = rd.significant_number(impact)

        warnings = ["Generic data used for impact calculation."]

        return impact, significant_figures, min_impacts, max_impacts, warnings

class DeviceTablet(EndUserDevice, ABC):
    NAME = "TABLET"
    def __init__(self, archetype=get_user_terminal_archetype(config["default_tablet"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)

class DeviceSmartphone(EndUserDevice, ABC):
    NAME = "SMARTPHONE"
    def __init__(self, archetype=get_user_terminal_archetype(config["default_smartphone"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)

class DeviceTelevision(EndUserDevice, ABC):
    NAME = "TELEVISION"
    def __init__(self, archetype=get_user_terminal_archetype(config["default_television"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)
        self.type = Boattribute(
            default=get_arch_value(archetype, 'type', 'default'),
            min=get_arch_value(archetype, 'type', 'min'),
            max=get_arch_value(archetype, 'type', 'max')
        )

    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        impact = float(get_impact_factor(item=self.NAME, impact_type=impact_type)[self.type.value]["impact"])
        min_impacts = float(get_impact_factor(item=self.NAME, impact_type=impact_type)[self.type.value]["impact"])
        max_impacts = float(get_impact_factor(item=self.NAME, impact_type=impact_type)[self.type.value]["impact"])
        significant_figures = rd.significant_number(impact)

        warnings = ["Generic data used for impact calculation."]

        return impact, significant_figures, min_impacts, max_impacts, warnings

class DeviceSmartWatch(EndUserDevice, ABC):
    NAME = "SMARTWATCH"
    def __init__(self, archetype=get_user_terminal_archetype(config["default_smartwatch"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)

class DeviceBox(EndUserDevice, ABC):
    NAME = "BOX"
    def __init__(self, archetype=get_user_terminal_archetype(config["default_box"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)

class DeviceUsbStick(EndUserDevice, ABC):
    NAME = "USB_STICK"
    def __init__(self, archetype=get_user_terminal_archetype(config["default_usb_stick"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)

class DeviceExternalSSD(EndUserDevice, ABC):
    NAME = "EXTERNAL_SSD"
    def __init__(self, archetype=get_user_terminal_archetype(config["default_external_ssd"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)

class DeviceExternalHDD(EndUserDevice, ABC):
    NAME = "EXTERNAL_HDD"
    def __init__(self, archetype=get_user_terminal_archetype(config["default_external_hdd"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)

class DeviceMonitor(EndUserDevice, ABC):
    NAME = "MONITOR"
    def __init__(self, archetype=get_user_terminal_archetype(config["default_monitor"]), **kwargs):
        super().__init__(archetype=archetype, **kwargs)


server_impact_by_model_description = "# ✔ ️Server impacts from model name\n" \
                                     "Retrieve the impacts of a given server archetype.\n" \
                                     "### Features\n\n" \
                                     "👄 Verbose\n\n" \
                                     "🔃 Auto-complete\n\n" \
                                     "🔨 Embedded\n\n" \
                                     "🔌 Usage\n\n" \
                                     "📋 Archetype: " \
                                     "Uses the classic server impacts router with a pre-registered archetype \n\n" \
                                     "⏬ Allocation"

server_impact_by_config_description = "# ✔️ Server impacts from configuration\n" \
                                      "Retrieve the impacts of a given server configuration.\n" \
                                      "### Features\n\n" \
                                      "👄 Verbose\n\n" \
                                      "🔃 Auto-complete\n\n" \
                                      "🔨 Embedded\n\n" \
                                      "🔌 Usage\n\n" \
                                      "* ⏺️  Given\n\n" \
                                      "* 📈 Modeled\n\n" \
                                      "📋 Archetype\n\n" \
                                      "⏬ Allocation"

all_archetype_servers = "# ✔️ Get all the available server archetype\n"
all_archetype_components = "# ✔️ Get all the available component archetype for a given component name\n"
all_archetype_user_terminals = "# ✔️ Get all the available user terminal archetype for a given user terminal name\n"
all_terminal_categories = "# ✔️ Get all the available user terminal router\n"
all_peripheral_categories = "# ✔️ Get all the available user peripheral router\n"
all_user_terminal_subcategories = "# ✔️ Get all the available user terminal subcategories\n"
all_default_usage_values = "# ✔️ Get all default usage values for a given user terminal category and subcategory\n"

get_archetype_config_desc = "# ✔️ Get the configuration of a given archetype\n"
get_instance_config = "# ✔️ Get the configuration of a given instance\n"

cpu_description = "# ✔ ️CPU impacts from configuration\n" \
                  "### Features\n\n" \
                  "👄 Verbose\n\n" \
                  "🔃 Auto-complete\n\n" \
                  "🔨 Embedded\n\n" \
                  "<h3>cpu<sub>manuf<sub><em>criteria</em></sub></sub> = ( " \
                  "cpu<sub>core<sub>units</sub></sub> x cpu<sub>diesize</sub> + 0," \
                  "491 ) x cpu<sub>manuf_die<sub><em>criteria</em></sub></sub> + " \
                  "cpu<sub>manuf_base<sub><em>criteria</em></sub></sub></h3> " \
                  "🔌 Usage\n\n" \
                  "* ⏺️  Given\n\n" \
                  "* 📈 Modeled\n\n" \
                  "⏬ Allocation"

ssd_description = "# ✔ ️SSD impacts from configuration\n" \
                  "### Features\n\n" \
                  "👄 Verbose\n\n" \
                  "🔃 Auto-complete\n\n" \
                  "🔨 Embedded\n\n" \
                  "<h3>ssd<sub>manuf<sub><em>criteria</em></sub></sub> =  ( ssd<sub>size</sub> " \
                  "ssd<sub>density</sub> ) x ssd<sub>manuf_die<sub><em>criteria</em></sub></sub> + " \
                  "ssd<sub>manuf_base<sub><em>criteria</em></sub></sub></h3>" \
                  "🔌 Usage\n\n" \
                  "* ⏺️  Given\n\n" \
                  "⏬ Allocation"

hdd_description = "# ✔ ️HDD impacts from configuration\n" \
                  "### Features\n\n" \
                  "👄 Verbose\n\n" \
                  "🔃 Auto-complete\n\n" \
                  "🔨 Embedded\n\n" \
                  "The impacts values are set by default" \
                  "🔌 Usage\n\n" \
                  "* ⏺️  Given\n\n" \
                  "⏬ Allocation"

ram_description = "# ✔️ RAM impacts from configuration\n" \
                  "### Features\n\n" \
                  "👄 Verbose\n\n" \
                  "🔃 Auto-complete\n\n" \
                  "🔨 Embedded\n\n" \
                  "<h3>ram<sub>manuf<sub><em>criteria</em></sub></sub> =( ram<sub>size</sub> " \
                  "/ ram<sub>density</sub> ) x ram<sub>manuf_die<sub><em>criteria</em></sub></sub> + " \
                  "ram<sub>manuf_base<sub><em>criteria</em></sub></sub> </h3> " \
                  "🔌 Usage\n\n" \
                  "* ⏺️  Given\n\n" \
                  "* 📈 Modeled\n\n" \
                  "⏬ Allocation"

motherboard_description = "# ✔ ️Motherboard impacts from configuration\n" \
                          "### Features\n\n" \
                          "👄 Verbose\n\n" \
                          "🔃 Auto-complete\n\n" \
                          "🔨 Embedded\n\n" \
                          "The impacts values are set by default" \
                          "🔌 Usage\n\n" \
                          "* ⏺️  Given\n\n" \
                          "⏬ Allocation"

power_supply_description = "# ✔ ️Power supply impacts from configuration\n" + \
                           "### Features\n\n" \
                           "👄 Verbose\n\n" \
                           "🔃 Auto-complete\n\n" \
                           "🔨 Embedded\n\n" \
                           "<h3>psu<sub>manuf<sub><em>criteria</em></sub></sub> = psu<sub>unit<sub>weight</sub></sub>" \
                           " x psu<sub>manuf_weight<sub><em>criteria</em></sub></sub></h3> " \
                           "🔌 Usage\n\n" \
                           "* ⏺️  Given : shouldn't be used\n\n" \
                           "⏬ Allocation"

case_description = "# ✔ ️Case impacts from configuration\n" \
                   "### Features\n\n" \
                   "👄 Verbose\n\n" \
                   "🔃 Auto-complete\n\n" \
                   "🔨 Embedded\n\n" \
                   "🔌 Usage\n\n" \
                   "* ⏺️  Given : when the enclosure consumes energy \n\n" \
                   "⏬ Allocation"

cloud_provider_description = "# ✔ ️Cloud instance impacts from provider, instance type and usage \n" \
                             "Retrieve the impacts of a given Cloud instance and usage.\n\n" \
                             "### Features\n\n" \
                             "📋 Provider \n\n" \
                             "Name of the cloud provider. You can retrieve the [list here](" \
                             "#/cloud_instance/server_get_all_cloud_providers).\n\n" \
                             "📋 Instance type \n\n" \
                             "Name of the chosen instance. You can retrieve the [list here](" \
                             "#/cloud/server_get_archetype_name_v1_cloud_all_aws_instances_get).\n\n" \
                             "👄 Verbose\n\n" \
                             "🔨 Embedded\n\n" \
                             "🔌 Usage \n\n" \
                             "* 📈 Modeled\n\n" \
                             "📋 Archetype : The configuration is set by the API, only usage is given by the user\n\n" \
                             "⏬ Allocation"

all_default_cloud_instances = "# ✔ ️Get all the available instances for a given Cloud provider\n" \
                              "📜 Return the name of all pre-registered instances for the Cloud provider"

all_default_cloud_providers = "# ✔ ️Get all the available Cloud providers\n" \
                              "📜 Return the names of all pre-registered Cloud providers"

country_code = "# ✔ ️Get all the available countries with their trigram code (*usage:{usage_location: 'FRA'}*)\n"
cpu_family = "# ✔ ️Get all the available cpu family in the API (*cpu:{family:'skylake'}*)\n"
cpu_model_range = "# ✔ ️Get all the available cpu family in the API (*cpu:{model_range:'xeon platinum'}*)\n"
ssd_manufacturer = "# ✔ ️Get all the available ssd manufacturer in the API (*ssd:{manufacturer:'samsung'}*)\n"
ram_manufacturer = "# ✔ ️Get all the available ram manufacturer in the API (*ram:{manufacturer:'samsung'}*)\n"
case_type = "# ✔ ️Get all the available case type in the API (*model:{case:'blade'}*)\n"
name_to_cpu = "# ✔ ️Complete a cpu attributes from a cpu name\n"
cpu_names = "# ✔ ️Get all the available cpu name in the API (*cpu:{name:'intel xeon platinum 8175m'}*)\n"
impacts_criteria = "# ✔ ️Get all the available criteria for the impacts calculation\n"


terminal_description = "# ✔ Terminal impacts\n" \
                          "### Features\n\n" \
                          "👄 Verbose\n\n" \
                          "🔃 Auto-complete\n\n" \
                          "🔨 Embedded\n\n" \
                          "The impacts values are fix" \
                          "🔌 Usage\n\n" \
                          "* ⏺️  Given\n\n" \
                          "* 📋 Archetype\n\n" \
                          "⏬ Allocation"


peripheral_description = "# ✔ Peripheral impacts\n" \
                          "### Features\n\n" \
                          "👄 Verbose\n\n" \
                          "🔃 Auto-complete\n\n" \
                          "🔨 Embedded\n\n" \
                          "The impacts values are fix" \
                          "🔌 Usage\n\n" \
                          "* ⏺️  Given\n\n" \
                          "* 📋 Archetype\n\n" \
                          "⏬ Allocation"



import boaviztapi.utils.roundit as rd
from boaviztapi import config
from boaviztapi.model.boattribute import Boattribute
from boaviztapi.model.component.component import Component, ComputedImpacts
from boaviztapi.model.impact import ImpactFactor
from boaviztapi.service.archetype import get_component_archetype, get_arch_value
from boaviztapi.service.factor_provider import get_impact_factor


class ComponentPowerSupply(Component):
    NAME = "POWER_SUPPLY"

    def __init__(self, archetype=get_component_archetype(config["default_power_supply"], "power_supply"), **kwargs):
        super().__init__(archetype=archetype, **kwargs)

        self.unit_weight = Boattribute(
            unit="kg",
            default=get_arch_value(archetype, 'unit_weight', 'default'),
            min=get_arch_value(archetype, 'unit_weight', 'min'),
            max=get_arch_value(archetype, 'unit_weight', 'max')
        )

    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        impact_factor = ImpactFactor(
            value=get_impact_factor(item='power_supply', impact_type=impact_type)['impact'],
            min=get_impact_factor(item='power_supply', impact_type=impact_type)['impact'],
            max=get_impact_factor(item='power_supply', impact_type=impact_type)['impact']
        )

        impact = self.__compute_impact_manufacture(impact_factor)
        sign_figures = rd.min_significant_figures(impact_factor.value)

        return impact.value, sign_figures, impact.min, impact.max, ["End of life is not included in the calculation"]

    def __compute_impact_manufacture(self, power_supply_impact: ImpactFactor) -> ImpactFactor:
        return ImpactFactor(
            value=self.unit_weight.value * power_supply_impact.value,
            min=self.unit_weight.min * power_supply_impact.min,
            max=self.unit_weight.max * power_supply_impact.max
        )

    def impact_use(self, impact_type: str, duration: float) -> ComputedImpacts:
        raise NotImplementedError



import os
from typing import Tuple

import pandas as pd

import boaviztapi.utils.roundit as rd
from boaviztapi import config, data_dir
from boaviztapi.model.boattribute import Boattribute
from boaviztapi.model.component.component import Component, ComputedImpacts
from boaviztapi.model.consumption_profile.consumption_profile import RAMConsumptionProfileModel
from boaviztapi.model.impact import ImpactFactor
from boaviztapi.service.archetype import get_arch_value, get_component_archetype
from boaviztapi.service.factor_provider import get_impact_factor
from boaviztapi.utils.fuzzymatch import fuzzymatch_attr_from_pdf


class ComponentRAM(Component):
    NAME = "RAM"

    _ram_df = pd.read_csv(os.path.join(data_dir, 'crowdsourcing/ram_manufacture.csv'))

    def __init__(self, archetype=get_component_archetype(config["default_ram"], "ram"), **kwargs):
        super().__init__(archetype=archetype, **kwargs)

        self.process = Boattribute(
            default=get_arch_value(archetype, 'process', 'default'),
            min=get_arch_value(archetype, 'process', 'min'),
            max=get_arch_value(archetype, 'process', 'max')
        )
        self.manufacturer = Boattribute(
            default=get_arch_value(archetype, 'manufacturer', 'default'),
            min=get_arch_value(archetype, 'manufacturer', 'min'),
            max=get_arch_value(archetype, 'manufacturer', 'max')
        )
        self.capacity = Boattribute(
            unit="GB",
            default=get_arch_value(archetype, 'capacity', 'default'),
            min=get_arch_value(archetype, 'capacity', 'min'),
            max=get_arch_value(archetype, 'capacity', 'max')
        )
        self.density = Boattribute(
            complete_function=self._complete_density,
            unit="GB/cm2",
            default=get_arch_value(archetype, 'density', 'default'),
            min=get_arch_value(archetype, 'density', 'min'),
            max=get_arch_value(archetype, 'density', 'max')
        )
        self.usage.avg_power.add_warning("value for one ram strip")


    # IMPACT COMPUTATION
    def impact_embedded(self, impact_type: str) -> ComputedImpacts:
        ram_die_impact, ram_impact = self.__get_impact_constants(impact_type)

        impact = self.__compute_impact_manufacture(ram_die_impact, ram_impact)

        sign_figures = self.__compute_significant_numbers(ram_die_impact.value, ram_impact.value)
        return impact.value, sign_figures, impact.min, impact.max, ["End of life is not included in the calculation"]

    def impact_use(self, impact_type: str, duration: float) -> ComputedImpacts:
        impact_factor = self.usage.elec_factors[impact_type]

        if not self.usage.avg_power.is_set():
            modeled_consumption = self.model_power_consumption()
            self.usage.avg_power.set_completed(
                modeled_consumption.value,
                min=modeled_consumption.min,
                max=modeled_consumption.max,
            )

        impact = ImpactFactor(
            value=impact_factor.value * (
                    self.usage.avg_power.value / 1000) * self.usage.use_time_ratio.value * duration,
            min=impact_factor.min * (
                    self.usage.avg_power.min / 1000) * self.usage.use_time_ratio.min * duration,
            max=impact_factor.max * (
                    self.usage.avg_power.max / 1000) * self.usage.use_time_ratio.max * duration
        )

        sig_fig = self.__compute_significant_numbers_usage(impact_factor.value)
        return impact.value, sig_fig, impact.min, impact.max, []

    def __compute_significant_numbers_usage(self, impact_factor: float) -> int:
        return rd.min_significant_figures(self.usage.avg_power.value, self.usage.use_time_ratio.value, impact_factor)

    def model_power_consumption(self,) -> ImpactFactor:
        self.usage.consumption_profile = RAMConsumptionProfileModel()
        self.usage.consumption_profile.compute_consumption_profile_model(ram_capacity=self.capacity.value)

        if type(self.usage.time_workload.value) in (float, int):
            self.usage.avg_power.set_completed(
                self.usage.consumption_profile.apply_consumption_profile(self.usage.time_workload.value))
        else:
            self.usage.avg_power.set_completed(
                self.usage.consumption_profile.apply_multiple_workloads(self.usage.time_workload.value))

        return ImpactFactor(
                value=rd.round_to_sigfig(self.usage.avg_power.value, 5),
                min=rd.round_to_sigfig(self.usage.avg_power.value, 5),
                max=rd.round_to_sigfig(self.usage.avg_power.value, 5)
        )

    def __get_impact_constants(self, impact_type: str) -> Tuple[ImpactFactor, ImpactFactor]:
        ram_die_impact = ImpactFactor(
            value=get_impact_factor(item='ram', impact_type=impact_type)['die_impact'],
            min=get_impact_factor(item='ram', impact_type=impact_type)['die_impact'],
            max=get_impact_factor(item='ram', impact_type=impact_type)['die_impact']
        )
        ram_impact = ImpactFactor(
            value=get_impact_factor(item='ram', impact_type=impact_type)['impact'],
            min=get_impact_factor(item='ram', impact_type=impact_type)['impact'],
            max=get_impact_factor(item='ram', impact_type=impact_type)['impact']
        )
        return ram_die_impact, ram_impact

    def __compute_significant_numbers(self, ram_die_impact: float, ram_impact: float) -> int:
        return rd.min_significant_figures(self.density.value, ram_die_impact, ram_impact)

    def __compute_impact_manufacture(self, ram_die_impact: ImpactFactor, ram_impact: ImpactFactor) -> ImpactFactor:
        return ImpactFactor(
            value=(self.capacity.value / self.density.value) * ram_die_impact.value + ram_impact.value,
            min=(self.capacity.min / self.density.max) * ram_die_impact.min + ram_impact.min,
            max=(self.capacity.max / self.density.min) * ram_die_impact.max + ram_impact.max
        )

    # COMPLETION
    def _complete_density(self):
        sub = self._ram_df

        if self.manufacturer.has_value():
            corrected_manufacturer = fuzzymatch_attr_from_pdf(self.manufacturer.value, "manufacturer", sub)
            sub = sub[sub['manufacturer'] == corrected_manufacturer]
            if corrected_manufacturer != self.manufacturer.value:
                self.manufacturer.set_changed(corrected_manufacturer)

        if self.process.has_value():
            sub = sub[sub['process'] == self.process.value]

        if len(sub) == 1:
            self.density.set_completed(
                float(sub['density']),
                source=str(sub['manufacturer'].iloc[0]),
                min=float(sub['density']),
                max=float(sub['density'])
            )

        elif (len(sub) == 0 or len(sub) == len(self._ram_df)) and self.density.has_value():
            return

        else:
            self.density.set_completed(
                float(sub['density'].mean()),
                source="Average of " + str(len(sub)) + " rows",
                min=float(sub['density'].min()),
                max=float(sub['density'].max())
            )


import dataclasses
import math
import os
from typing import Dict, Optional, List, Tuple, Union

import numpy as np
import pandas as pd
from scipy.optimize import curve_fit

import boaviztapi.utils.fuzzymatch as fuzzymatch
from boaviztapi import config, data_dir
from boaviztapi.dto.usage.usage import WorkloadTime
from boaviztapi.model.boattribute import Boattribute, Status
from boaviztapi.service.archetype import get_component_archetype, get_arch_value

fuzzymatch.pandas()

_cpu_profile_consumption_df = pd.read_csv(os.path.join(data_dir, 'consumption_profile/cpu/cpu_profile.csv'))

class ConsumptionProfileModel:
    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value


class RAMConsumptionProfileModel(ConsumptionProfileModel):
    ram_electrical_factor_per_go = 0.284

    def __init__(self, archetype=get_component_archetype(config["default_ram"], "ram").get("consumption_profile")):
        self.workloads = Boattribute(
            unit="workload_rate:W"
        )
        self.params = Boattribute()

    def compute_consumption_profile_model(self, ram_capacity) -> int:
        self.params.value = {'a': self.ram_electrical_factor_per_go * ram_capacity}
        self.params.status = Status.COMPLETED
        self.params.source = f"(ram_electrical_factor_per_go : {self.ram_electrical_factor_per_go}) * (" \
                             f"ram_capacity: {ram_capacity}) "
        return self.params.value

    def apply_consumption_profile(self, load_percentage: float) -> float:
        return self.params.value['a']

    def apply_multiple_workloads(self, time_workload: List[WorkloadTime]) -> float:
        total = 0
        for workload in time_workload:
            total += (workload.time_percentage / 100) * self.apply_consumption_profile(workload.load_percentage)
        return total


class CPUConsumptionProfileModel(ConsumptionProfileModel):
    _TDP_RATIOS_WORKLOAD = [0, 10, 50, 100]
    _TDP_RATIOS = [0.12, 0.32, 0.75, 1.02]

    _DEFAULT_MODEL_BOUNDS = (
        [0, 0, 0, -math.inf],
        [math.inf, math.inf, math.inf, math.inf]
    )
    _MODEL_PARAM_NAME = ['a', 'b', 'c', 'd']

    def __init__(self, archetype=get_component_archetype(config["default_cpu"], "cpu").get("CONSUMPTION_PROFILE")):
        self.workloads = Boattribute(
            unit="workload_rate:W"
        )
        self.params = Boattribute(
            default=get_arch_value(archetype, 'params', 'default'),
            min=get_arch_value(archetype, 'params', 'min'),
            max=get_arch_value(archetype, 'params', 'max')
        )

    @property
    def list_workloads(self) -> Tuple[List[float], List[float]]:
        load = [item.load_percentage for item in self.workloads.value]
        power = [item.power_watt for item in self.workloads.value]
        return load, power

    def apply_consumption_profile(self, load_percentage: float) -> float:
        return self.__log_model(
            load_percentage,
            self.params.value['a'],
            self.params.value['b'],
            self.params.value['c'],
            self.params.value['d']
        )

    def apply_multiple_workloads(self, time_workload: List[WorkloadTime]) -> float:
        total = 0
        for workload in time_workload:
            total += (workload.time_percentage / 100) * self.apply_consumption_profile(workload.load_percentage)
        return total

    def compute_consumption_profile_model(self,
                                          cpu_manufacturer: str = None,
                                          cpu_model_range: str = None,
                                          cpu_tdp: int = None) -> Union[Dict[str, float], None]:
        model = self.lookup_consumption_profile(cpu_manufacturer, cpu_model_range)
        if self.workloads.is_set():
            model = self.__compute_model_adaptation(base_model=model or self.params.default)
            self.params.set_completed(model, source="From workload")

        elif cpu_tdp is not None:
            model = self.__compute_model_adaptation_with_tdp(
                base_model=model or self.params.default,
                cpu_tdp=cpu_tdp
            )
            self.params.set_completed(model, source="From TDP")

        elif cpu_model_range is not None:
            self.params.set_completed(model, source="From CPU model range")

        return self.params.value

    def __compute_model_adaptation_with_tdp(self, base_model: Dict[str, float], cpu_tdp: float) -> Dict[str, float]:
        @dataclasses.dataclass
        class _TDPWorkloadPower:
            load_percentage: float = None
            power_watt: float = None
        self.workloads.set_completed([
            _TDPWorkloadPower(load_percentage=w, power_watt=cpu_tdp * r)
            for w, r in zip(self._TDP_RATIOS_WORKLOAD, self._TDP_RATIOS)
        ])
        return self.__compute_model_adaptation(base_model=base_model)

    def __compute_model_adaptation(self, base_model: Dict[str, float]) -> Dict[str, float]:
        base_model_list = self.__model_dict_to_list(base_model)
        bounds = self.__adapt_model_bounds(base_model_list)
        x_data, y_data = self.list_workloads
        popt, _ = curve_fit(f=self.__log_model,
                            xdata=x_data,
                            ydata=y_data,
                            p0=base_model_list,
                            bounds=bounds)
        return self.__model_list_to_dict(popt.tolist())

    def __adapt_model_bounds(self, base_model_list: List[float]) -> Tuple[List[float], List[float]]:
        default_lower_bounds, default_upper_bounds = self._DEFAULT_MODEL_BOUNDS
        lower_bounds, upper_bounds = [], []
        for lower_b, upper_b, model_param in zip(default_lower_bounds, default_upper_bounds, base_model_list):
            lower_bounds.append(max(lower_b, model_param - abs(0.5 * model_param)))
            upper_bounds.append(min(upper_b, model_param + abs(1.5 * model_param)))
        return lower_bounds, upper_bounds

    def __model_dict_to_list(self, model: Dict[str, float]) -> List[float]:
        return [model[param_name] for param_name in self._MODEL_PARAM_NAME]

    def __model_list_to_dict(self, model: List[float]) -> Dict[str, float]:
        return {param_name: param for param_name, param in zip(self._MODEL_PARAM_NAME, model)}

    @staticmethod
    def __log_model(x: float, a: float, b: float, c: float, d: float) -> float:
        return a * np.log(b * (x + c)) + d

    @staticmethod
    def lookup_consumption_profile(
            cpu_manufacturer: str = None,
            cpu_model_range: str = None
    ) -> Optional[Dict[str, float]]:
        sub = _cpu_profile_consumption_df

        if cpu_manufacturer is not None:
            tmp = sub[sub['manufacturer'].fuzzymatch(cpu_manufacturer)]
            if len(tmp) > 0:
                sub = tmp.copy()

        if cpu_model_range is not None:
            tmp = sub[sub['model_range'].fuzzymatch(cpu_model_range)]
            if len(tmp) > 0:
                sub = tmp.copy()

        if len(sub) == 1:
            row = sub.iloc[0]
            return {
                'a': row.a,
                'b': row.b,
                'c': row.c,
                'd': row.d,
            }
